import logging
import math
from typing import Dict, Optional, Tuple

from player_performance_ratings.data_structures import MatchPlayer, PlayerRating, PreMatchTeamRating, \
    PreMatchPlayerRating, \
    PlayerRatingChange, Team, MatchTeam, TeamRatingChange
from player_performance_ratings.ratings.rating_calculators.performance_predictor import \
    RatingDifferencePerformancePredictor, \
    PerformancePredictor
from player_performance_ratings.ratings.rating_calculators.start_rating_generator import \
    StartRatingGenerator

MATCH_CONTRIBUTION_TO_SUM_VALUE = 1
EXPECTED_MEAN_CONFIDENCE_SUM = 30


class MatchRatingGenerator():

    def __init__(self,
                 start_rating_generator: Optional[StartRatingGenerator] = None,
                 performance_predictor: Optional[PerformancePredictor] = None,
                 rating_change_multiplier: float = 50,
                 confidence_days_ago_multiplier: float = 0.06,
                 confidence_max_days: int = 90,
                 confidence_value_denom: float = 140,
                 confidence_max_sum: float = 150,
                 confidence_weight: float = 0.9,
                 min_rating_change_multiplier_ratio: float = 0.1,
                 league_rating_change_update_threshold: float = 100,
                 league_rating_adjustor_multiplier: float = 0.05,
                 team_id_change_confidence_sum_decrease: float = 3
                 ):

        """
        :param start_rating_generator:
            Class object for generating start rating (the rating a player receives in its very first match).

        :param performance_predictor:
            Class object for predicting performance of a player in a match.
        This is used to adjust for the current quality of opponents faced or other contexts dependant on the class used.

        :param rating_change_multiplier:
            The base multiplier for how much a player's rating changes after a match.
            Equivalent of the K-factor in Elo.


        :param confidence_days_ago_multiplier:
            Determinutes how much the confidence_sum is affected by how long ago the last match was.
            If a player has not played in a long time, the confidence in a players rating declines.
            This means his future rating changes faster. A higher confidence_days_ago_multiplier  therefore leads to more volatile future rating changes.

        :param confidence_max_days:
            A players confidence_sum decreases by confidence_days_ago_multiplier * min(confidence_max_days, days_ago_since_last_match).
            confidence_max_days therefore puts a limit on how much a players confidence_sum can decrease.
            This ensures a non-linear relationship between inactivity and our confidence in a players rating.

        :param confidence_value_denom:
            Higher confidence_value_denom results in a more volatile rating change multipler based on the confidence_sum.

        :param confidence_max_sum:
            The maximum confidence_sum a player can have.
             If confidence_max_sum is very high, a player with many historical games, can be inactive for a while and still have a high confidence_sum.
             Thus a lower confidence_max_sum increases the speed at which a players confidence_sum decreases when inactive.


        :param min_rating_change_multiplier_ratio:
            To ensure that a player's rating changes at least a little bit after a match,
            min_rating_change_multiplier_ratio ensures that the applied_rating_change_multiplier is at least a certain ratio of the rating_change_multiplier.

        :param confidence_weight:
            Determines how much the applied rating change multiplier is affected by the confidence_sum.
            The confidence_sum being how much data (and how recent) we have on a player.
            If the confidence_ratio is 0, the applied rating_change_multipler is equal to the rating_change_multiplier.
            The higher ratio it is, the more volatile the applied rating_change_multiplier is based on the data we have on a player.
            Recommended is between 0.7 and 1.0

        :param league_rating_change_update_threshold:
            All players within a league gets their rating updated when the sum of all rating changes within a league exceeds this threshold.
            A lower value means that the ratings within a league are updated more often but is more computationally expensive.

        :param league_rating_adjustor_multiplier:
            The amount of which the players within a league gets their rating updated with, is league_rating_adjustor_multiplier * mean_rating_change.
            A higher value makes the performance of other players within a league have a larger impact on a players rating - even if the player hasn't played a match.

        """

        self.confidence_weight = confidence_weight
        self.confidence_days_ago_multiplier = confidence_days_ago_multiplier
        self.confidence_value_denom = confidence_value_denom
        self.min_rating_change_multiplier_ratio = min_rating_change_multiplier_ratio
        self.rating_change_multiplier = rating_change_multiplier
        self.confidence_max_sum = confidence_max_sum
        self.league_rating_adjustor_multiplier = league_rating_adjustor_multiplier
        self.league_rating_change_update_threshold = league_rating_change_update_threshold
        self.confidence_max_days = confidence_max_days
        self.player_ratings: Dict[str, PlayerRating] = {}
        self._league_rating_changes: dict[Optional[str], float] = {}
        self._league_rating_changes_count: dict[str, float] = {}
        self.performance_predictor = performance_predictor or RatingDifferencePerformancePredictor()
        self.start_rating_generator = start_rating_generator or StartRatingGenerator()
        self.team_id_change_confidence_sum_decrease = team_id_change_confidence_sum_decrease

        self._teams: dict[str, Team] = {}

    def generate_pre_match_team_rating(self, day_number: int, match_team: MatchTeam) -> PreMatchTeamRating:

        self._teams[match_team.id] = Team(
            id=match_team.id,
            last_match_day_number=day_number,
            player_ids=[p.id for p in match_team.players]
        )

        existing_pre_match_player_ratings, pre_match_player_rating_values, new_players = self._get_pre_match_player_ratings_and_new_players(
            team=match_team)

        new_player_pre_match_ratings = self._generate_new_player_pre_match_ratings(day_number=day_number,
                                                                                   new_players=new_players,
                                                                                   team_pre_match_player_ratings=existing_pre_match_player_ratings)
        if len(new_player_pre_match_ratings) > 0 and len(existing_pre_match_player_ratings) > 0:

            pre_match_player_ratings = []
            for match_player in match_team.players:

                pre_match_player = [p for p in new_player_pre_match_ratings if p.id == match_player.id]
                if len(pre_match_player) == 0:
                    pre_match_player = [p for p in existing_pre_match_player_ratings if p.id == match_player.id]

                pre_match_player_ratings.append(pre_match_player[0])

        elif len(new_player_pre_match_ratings) == 0:
            pre_match_player_ratings = existing_pre_match_player_ratings
        else:
            pre_match_player_ratings = new_player_pre_match_ratings

        # pre_match_player_ratings += new_player_pre_match_ratings
        pre_match_team_rating_value = self._generate_pre_match_team_rating_value(
            pre_match_player_ratings=pre_match_player_ratings)
        pre_match_team_rating_projected_value = self._generate_pre_match_team_rating_projected_value(
            pre_match_player_ratings=pre_match_player_ratings)

        return PreMatchTeamRating(
            id=match_team.id,
            players=pre_match_player_ratings,
            rating_value=pre_match_team_rating_value,
            projected_rating_value=pre_match_team_rating_projected_value,
            league=match_team.league
        )

    def _get_pre_match_player_ratings_and_new_players(self, team: MatchTeam) -> Tuple[
        list[PreMatchPlayerRating], list[float], list[MatchPlayer]]:

        pre_match_player_ratings = []
        player_count = 0

        new_match_entities = []
        pre_match_player_ratings_values = []

        for match_player in team.players:
            if match_player.id in self.player_ratings:

                player_rating = self._get_rating_by_id(id=match_player.id)

                pre_match_player_rating = PreMatchPlayerRating(
                    id=match_player.id,
                    rating_value=player_rating.rating_value,
                    match_performance=match_player.performance,
                    games_played=player_rating.games_played,
                    league=match_player.league,
                    position=match_player.position,
                    other=match_player.others
                )

            else:
                new_match_entities.append(match_player)
                continue

            player_count += self._get_rating_by_id(match_player.id).games_played

            pre_match_player_ratings.append(pre_match_player_rating)
            pre_match_player_ratings_values.append(pre_match_player_rating.rating_value)

        return pre_match_player_ratings, pre_match_player_ratings_values, new_match_entities

    def _generate_new_player_pre_match_ratings(self, day_number: int, new_players: list[MatchPlayer],
                                               team_pre_match_player_ratings: list[PreMatchPlayerRating]) -> list[
        PreMatchPlayerRating]:

        pre_match_player_ratings = []

        for match_player in new_players:
            id = match_player.id

            rating_value = self.start_rating_generator.generate_rating_value(
                day_number=day_number,
                match_player=match_player,
                team_pre_match_player_ratings=team_pre_match_player_ratings,
            )

            self.player_ratings[match_player.id] = PlayerRating(
                id=id,
                rating_value=rating_value
            )

            pre_match_player_rating = PreMatchPlayerRating(
                id=id,
                rating_value=rating_value,
                match_performance=match_player.performance,
                games_played=self.player_ratings[match_player.id].games_played,
                league=match_player.league,
                position=match_player.position,
                other=match_player.others
            )

            pre_match_player_ratings.append(pre_match_player_rating)

        return pre_match_player_ratings

    def _generate_pre_match_team_rating_projected_value(self,
                                                        pre_match_player_ratings: list[PreMatchPlayerRating]) -> float:
        team_rating = sum(
            player.rating_value * player.match_performance.projected_participation_weight for player in
            pre_match_player_ratings)
        sum_participation_weight = sum(
            player.match_performance.projected_participation_weight for player in pre_match_player_ratings)

        return team_rating / sum_participation_weight if sum_participation_weight > 0 else 0

    def _generate_pre_match_team_rating_value(self, pre_match_player_ratings: list[PreMatchPlayerRating]) -> Optional[
        float]:
        if len(pre_match_player_ratings) > 0 and pre_match_player_ratings[
            0].match_performance.participation_weight is None:
            return None
        team_rating = sum(
            player.rating_value * player.match_performance.participation_weight for player in pre_match_player_ratings)
        sum_participation_weight = sum(
            player.match_performance.participation_weight for player in pre_match_player_ratings)

        return team_rating / sum_participation_weight if sum_participation_weight > 0 else 0

    def generate_rating_change(self,
                               day_number: int,
                               pre_match_team_rating: PreMatchTeamRating,
                               pre_match_opponent_team_rating: PreMatchTeamRating,
                               ) -> TeamRatingChange:

        player_rating_changes = []
        sum_participation_weight = 0
        sum_predicted_performance = 0
        sum_performance_value = 0
        sum_rating_change = 0

        for pre_player_rating in pre_match_team_rating.players:

            predicted_performance = self.performance_predictor.predict_performance(
                player_rating=pre_player_rating,
                opponent_team_rating=pre_match_opponent_team_rating,
                team_rating=pre_match_team_rating
            )

            applied_rating_change_multiplier = self._calculate_applied_rating_change_multiplier(
                player_id=pre_player_rating.id, team_id=pre_match_team_rating.id)
            performance_difference = pre_player_rating.match_performance.performance_value - predicted_performance
            rating_change_value = performance_difference * applied_rating_change_multiplier * pre_player_rating.match_performance.participation_weight
            if math.isnan(rating_change_value):
                raise ValueError(f"rating_change_value is nan for {pre_player_rating.id}")

            player_rating_change = PlayerRatingChange(
                id=pre_player_rating.id,
                predicted_performance=predicted_performance,
                participation_weight=pre_player_rating.match_performance.participation_weight,
                performance=pre_player_rating.match_performance.performance_value,
                rating_change_value=rating_change_value,
                league=pre_player_rating.league,
                day_number=day_number,
                pre_match_rating_value=pre_player_rating.rating_value
            )
            player_rating_changes.append(player_rating_change)
            sum_predicted_performance += player_rating_change.predicted_performance * pre_player_rating.match_performance.participation_weight
            sum_performance_value += pre_player_rating.match_performance.performance_value * pre_player_rating.match_performance.participation_weight
            sum_rating_change += player_rating_change.rating_change_value * pre_player_rating.match_performance.participation_weight
            sum_participation_weight += pre_player_rating.match_performance.participation_weight

        predicted_performance = sum_predicted_performance / sum_participation_weight if sum_participation_weight > 0 else 0
        performance = sum_performance_value / sum_participation_weight if sum_participation_weight > 0 else 0

        return TeamRatingChange(
            players=player_rating_changes,
            id=pre_match_team_rating.id,
            rating_change_value=sum_rating_change,
            predicted_performance=predicted_performance,
            pre_match_projected_rating_value=pre_match_team_rating.rating_value,
            league=pre_match_team_rating.league,
            performance=performance
        )

    def update_rating_by_team_rating_change(self,
                                            team_rating_change: TeamRatingChange,
                                            opponent_team_rating_change: TeamRatingChange
                                            ) -> None:

        for player_rating_change in team_rating_change.players:
            id = player_rating_change.id
            self.player_ratings[id].confidence_sum = self._calculate_post_match_confidence_sum(
                entity_rating=self.player_ratings[id],
                day_number=player_rating_change.day_number,
                particpation_weight=player_rating_change.participation_weight
            )

            self.player_ratings[id].rating_value += player_rating_change.rating_change_value
            self.player_ratings[id].games_played += player_rating_change.participation_weight
            self.player_ratings[id].last_match_day_number = player_rating_change.day_number
            self.player_ratings[id].most_recent_team_id = team_rating_change.id

            self.start_rating_generator.update_players_to_leagues(rating_change=player_rating_change)
            if player_rating_change.league != opponent_team_rating_change.league:
                self._update_player_ratings_from_league_changes(rating_change=player_rating_change)

    def _update_player_ratings_from_league_changes(self,
                                                   rating_change: PlayerRatingChange
                                                   ):

        league = rating_change.league

        if league not in self._league_rating_changes:
            self._league_rating_changes[rating_change.league] = 0
            self._league_rating_changes_count[league] = 0

        rating_change_value = rating_change.rating_change_value
        self._league_rating_changes[rating_change.league] += rating_change_value
        self._league_rating_changes_count[league] += 1

        if self._league_rating_changes[league] > abs(self.league_rating_change_update_threshold):
            for player_id in self.start_rating_generator._league_to_player_ids[league]:
                self.player_ratings[
                    player_id].rating_value += self._league_rating_changes[
                                                   league] * self.league_rating_adjustor_multiplier

            self._league_rating_changes[league] = 0

    def _calculate_applied_rating_change_multiplier(self,
                                                    player_id: str,
                                                    team_id: str,
                                                    ) -> float:
        if self.player_ratings[player_id].most_recent_team_id and self.player_ratings[
            player_id].most_recent_team_id != team_id:
            self.player_ratings[player_id].confidence_sum -= self.team_id_change_confidence_sum_decrease

        min_applied_rating_change_multiplier = self.rating_change_multiplier * self.min_rating_change_multiplier_ratio
        confidence_change_multiplier = self.rating_change_multiplier * ((EXPECTED_MEAN_CONFIDENCE_SUM -
                                                                         self.player_ratings[
                                                                             player_id].confidence_sum) / self.confidence_value_denom + 1)
        applied_rating_change_multiplier = confidence_change_multiplier * self.confidence_weight + (
                1 - self.confidence_weight) * self.rating_change_multiplier

        # net_certain_sum_value = self.player_ratings[
        #                              player_id].confidence_sum - 20
        #  certain_factor = -(1 / (1 + math.exp(-net_certain_sum_value / 14)) - 0.5) * 2 + 1
        #   certain_multiplier = certain_factor * self.rating_change_multiplier
        #  multiplier = certain_multiplier * self.confidence_weight + (
        #          1 - self.confidence_weight) * self.rating_change_multiplier

        #   min_rating_change_multiplier = self.rating_change_multiplier * self.min_rating_change_multiplier_ratio
        return max(min_applied_rating_change_multiplier, applied_rating_change_multiplier)

    #     return max(min_applied_rating_change_multiplier, applied_rating_change_multiplier)

    def _calculate_post_match_confidence_sum(self,
                                             entity_rating: PlayerRating,
                                             day_number: int,
                                             particpation_weight: float
                                             ) -> float:
        days_ago = self._calculate_days_ago_since_last_match(last_match_day_number=entity_rating.last_match_day_number,
                                                             day_number=day_number)
        confidence_sum_value = -min(days_ago,
                                    self.confidence_max_days) * self.confidence_days_ago_multiplier + entity_rating.confidence_sum + \
                               MATCH_CONTRIBUTION_TO_SUM_VALUE * particpation_weight

        return max(0.0, min(confidence_sum_value, self.confidence_max_sum))

    def _calculate_days_ago_since_last_match(self, last_match_day_number, day_number: int) -> float:
        match_day_number = day_number
        if last_match_day_number is None:
            return 0.0

        return match_day_number - last_match_day_number

    def _get_rating_by_id(self, id: str):
        if id not in self.player_ratings:
            raise KeyError(f"{id} not in player_ratings")
        return self.player_ratings[id]

    @property
    def teams(self) -> dict[str, Team]:
        return self._teams



import logging
from typing import Optional, List

import numpy as np
import pandas as pd
from lightgbm import LGBMRegressor

from player_performance_ratings.predictor import Predictor
from player_performance_ratings.transformation.base_transformer import BaseTransformer


class NetOverPredictedTransformer(BaseTransformer):

    def __init__(self,
                 predictor: Predictor,
                 features: list[str],
                 prefix: str = "net_over_",
                 ):
        super().__init__(features=features)
        self.prefix = prefix
        self._predictor = predictor
        self._features_out = []
        new_feature_name = self.prefix + self._predictor.pred_column
        self._features_out.append(new_feature_name)
        if self.prefix is "":
            raise ValueError("Prefix must not be empty")

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        self._predictor.train(df, estimator_features=self.features)
        return self.transform(df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = self._predictor.add_prediction(df)
        new_feature_name = self.prefix + self._predictor.pred_column
        df = df.assign(**{new_feature_name: df[self._predictor.target] - df[self._predictor.pred_column]})
        df = df.drop(columns=[self._predictor.pred_column])

        return df

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class SklearnEstimatorImputer(BaseTransformer):

    def __init__(self, features: list[str], target_name: str, estimator: Optional[LGBMRegressor] = None):
        super().__init__(features=features)
        self.estimator = estimator or LGBMRegressor()
        self.target_name = target_name

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        self.estimator.fit(df[self.features], df[self.target_name])
        df = df.assign(**{
            f'imputed_col_{self.target_name}': self.estimator.predict(df[self.features])
        })
        df[self.target_name] = df[self.target_name].fillna(df[f'imputed_col_{self.target_name}'])
        return df.drop(columns=[f'imputed_col_{self.target_name}'])

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(**{
            f'imputed_col_{self.target_name}': self.estimator.predict(df[self.features])
        })
        df[self.target_name] = df[self.target_name].fillna(df[f'imputed_col_{self.target_name}'])
        return df.drop(columns=[f'imputed_col_{self.target_name}'])

    @property
    def features_out(self) -> list[str]:
        return [self.target_name]


class PartialStandardScaler(BaseTransformer):

    def __init__(self,
                 features: list[str],
                 ratio: float = 0.55,
                 target_mean: float = 0.5,
                 max_value: float = 2,
                 prefix: str = ""
                 ):
        super().__init__(features=features)
        self.ratio = ratio
        self.target_mean = target_mean
        self.prefix = prefix
        self.max_value = max_value

        self._features_mean = {}
        self._features_std = {}
        self._features_out = []
        for feature in self.features:
            self._features_out.append(self.prefix + feature)

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        for feature in self.features:
            self._features_mean[feature] = df[feature].replace([np.inf, -np.inf], np.nan).mean()
            self._features_std[feature] = df[feature].replace([np.inf, -np.inf], np.nan).std()

        return self.transform(df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        for feature in self.features:
            new_feature = self.prefix + feature
            df = df.assign(**{new_feature: ((df[feature] - self._features_mean[feature]) / self._features_std[
                feature]) * self.ratio + self.target_mean})
            df[new_feature] = df[new_feature].clip(-self.max_value + self.target_mean, self.max_value)
        return df

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class MinMaxTransformer(BaseTransformer):

    def __init__(self,
                 features: list[str],
                 quantile: float = 0.98,
                 multiply_align: bool = False,
                 add_align: bool = True,
                 prefix: str = ""
                 ):
        super().__init__(features=features)
        self.quantile = quantile
        self.prefix = prefix
        self._trained_mean_values = {}
        self._min_values = {}
        self._max_values = {}
        self.multiply_align = multiply_align
        self.add_align = add_align

        if self.quantile < 0 or self.quantile > 1:
            raise ValueError("quantile must be between 0 and 1")

        self._features_out = []

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        for feature in self.features:
            self._min_values[feature] = df[feature].quantile(1 - self.quantile)
            self._max_values[feature] = df[feature].quantile(self.quantile)

            if self._min_values[feature] == self._max_values[feature]:
                raise ValueError(
                    f"Feature {feature} has the same min and max value. This can lead to division by zero. "
                    f"This feature is not suited for MinMaxTransformer")

            df[self.prefix + feature] = (df[feature] - self._min_values[feature]) / (
                    self._max_values[feature] - self._min_values[feature])
            df[self.prefix + feature].clip(0, 1, inplace=True)
            self._trained_mean_values[feature] = df[self.prefix + feature].mean()
            if self.multiply_align:
                df[self.prefix + feature] = df[self.prefix + feature] * 0.5 / self._trained_mean_values[feature]
            if self.add_align:
                df[self.prefix + feature] = df[self.prefix + feature] + 0.5 - self._trained_mean_values[feature]

            self._features_out.append(self.prefix + feature)

        return df

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        for feature in self.features:
            df[self.prefix + feature] = (df[feature] - self._min_values[feature]) / (
                    self._max_values[feature] - self._min_values[feature])
            df[self.prefix + feature].clip(0, 1, inplace=True)
            if self.multiply_align:
                df[self.prefix + feature] = df[self.prefix + feature] * 0.5 / self._trained_mean_values[feature]
            if self.add_align:
                df[self.prefix + feature] = df[self.prefix + feature] + 0.5 - self._trained_mean_values[feature]

        return df

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class DiminishingValueTransformer(BaseTransformer):

    def __init__(self,
                 features: List[str],
                 cutoff_value: Optional[float] = None,
                 quantile_cutoff: float = 0.93,
                 excessive_multiplier: float = 0.8,
                 reverse: bool = False,
                 ):
        super().__init__(features=features)
        self.cutoff_value = cutoff_value
        self.excessive_multiplier = excessive_multiplier
        self.quantile_cutoff = quantile_cutoff
        self.reverse = reverse
        self._feature_cutoff_value = {}

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:

        for feature_name in self.features:
            if self.cutoff_value is None:
                if self.reverse:
                    self._feature_cutoff_value[feature_name] = df[feature_name].quantile(1 - self.quantile_cutoff)
                else:
                    self._feature_cutoff_value[feature_name] = df[feature_name].quantile(self.quantile_cutoff)
            else:
                self._feature_cutoff_value[feature_name] = self.cutoff_value

        return self.transform(df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        for feature_name in self.features:
            cutoff_value = self._feature_cutoff_value[feature_name]
            if self.reverse:
                df = df.assign(**{
                    feature_name: lambda x: np.where(
                        x[feature_name] <= cutoff_value,
                        - (cutoff_value - df[feature_name]) * self.excessive_multiplier + cutoff_value,
                        x[feature_name]

                    )
                })
            else:

                df = df.assign(**{
                    feature_name: lambda x: np.where(
                        x[feature_name] >= cutoff_value,
                        (x[feature_name] - cutoff_value).clip(lower=0) * self.excessive_multiplier + cutoff_value,
                        x[feature_name]
                    )
                })

            df = df.assign(**{feature_name: df[feature_name].fillna(df[feature_name])})
        return df

    @property
    def features_out(self) -> list[str]:
        return self.features


class SymmetricDistributionTransformer(BaseTransformer):

    def __init__(self,
                 features: List[str],
                 granularity: Optional[list[str]] = None,
                 skewness_allowed: float = 0.15,
                 max_iterations: int = 50,
                 min_excessive_multiplier: float = 0.04,
                 prefix: str = "symmetric_"
                 ):
        super().__init__(features=features)
        self.granularity = granularity
        self.skewness_allowed = skewness_allowed
        self.max_iterations = max_iterations
        self.min_excessive_multiplier = min_excessive_multiplier
        self.prefix = prefix

        self._diminishing_value_transformer = {}

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:

        if self.granularity:
            df = df.assign(__concat_granularity=df[self.granularity].astype(str).agg('_'.join, axis=1))

        for feature in self.features:
            if df[feature].min() == df[feature].max():
                raise ValueError(f"SymmetricDistributionTransformer: {feature} has the same min and max value.")
            self._diminishing_value_transformer[feature] = {}
            if self.granularity:

                unique_values = df["__concat_granularity"].unique()
                for unique_value in unique_values:
                    rows = df[df["__concat_granularity"] == unique_value]

                    self._fit(rows=rows, feature=feature, granularity_value=unique_value)

            else:
                self._fit(rows=df, feature=feature, granularity_value=None)

        return self.transform(df)

    def _fit(self, rows: pd.DataFrame, feature: str, granularity_value: Optional[str]) -> None:

        skewness = rows[feature].skew()
        excessive_multiplier = 0.8
        quantile_cutoff = 0.95

        iteration = 0
        while abs(skewness) > self.skewness_allowed and len(
                rows) > 10 and iteration < self.max_iterations:

            if skewness < 0:
                reverse = True
            else:
                reverse = False

            self._diminishing_value_transformer[feature][granularity_value] = DiminishingValueTransformer(
                features=[feature],
                reverse=reverse,
                excessive_multiplier=excessive_multiplier,
                quantile_cutoff=quantile_cutoff)
            transformed_rows = self._diminishing_value_transformer[feature][granularity_value].fit_transform(rows)
            new_excessive_multiplier = excessive_multiplier * 0.94
            if new_excessive_multiplier < self.min_excessive_multiplier:
                break
            excessive_multiplier = new_excessive_multiplier
            next_quantile_cutoff = quantile_cutoff * 0.994
            if transformed_rows[feature].quantile(next_quantile_cutoff) > transformed_rows[feature].min():
                quantile_cutoff = next_quantile_cutoff
            iteration += 1
            skewness = transformed_rows[feature].skew()

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        if self.granularity:
            df = df.assign(__concat_granularity=df[self.granularity].apply(lambda x: "_".join(x), axis=1))

        for feature in self.features:
            out_feature = self.prefix + feature
            if self.granularity:

                unique_values = df["__concat_granularity"].unique()
                if len(unique_values) > 100:
                    logging.warning(
                        f"SymmetricDistributionTransformer: {feature} has more than 100 unique values."
                        f" This can lead to long runtimes. Consider setting a lower granularity")

                if len(self._diminishing_value_transformer[feature]) == 0:
                    df[out_feature] = df[feature]
                else:

                    for unique_value in unique_values:
                        rows = df[df["__concat_granularity"] == unique_value]

                        if unique_value in self._diminishing_value_transformer[feature]:
                            rows = self._diminishing_value_transformer[feature][unique_value].transform(rows)

                        df.loc[df["__concat_granularity"] == unique_value, out_feature] = rows[feature]

            else:
                if None in self._diminishing_value_transformer[feature]:
                    df = self._diminishing_value_transformer[feature][None].transform(df)

        if '__concat_granularity' in df.columns:
            df = df.drop(columns=["__concat_granularity"])
        return df

    @property
    def features_out(self) -> list[str]:
        return [self.prefix + feature for feature in self.features]


class GroupByTransformer(BaseTransformer):

    def __init__(self,
                 features: list[str],
                 granularity: list[str],
                 agg_func: str = "mean",
                 prefix: str = "mean_",
                 ):
        super().__init__(features=features)
        self.granularity = granularity
        self.agg_func = agg_func
        self.prefix = prefix
        self._features_out = []
        self._grouped = None
        for feature in self.features:
            self._features_out.append(self.prefix + feature)

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        self._grouped = (df
                         .groupby(self.granularity)[self.features]
                         .agg(self.agg_func)
                         .reset_index()
                         .rename(columns={feature: self.prefix + feature for feature in self.features})
                         )

        return self.transform(df=df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        return df.merge(self._grouped, on=self.granularity, how='left')

    @property
    def features_out(self) -> list[str]:
        return self._features_out



from abc import abstractmethod, ABC
from typing import Optional

import pandas as pd
from player_performance_ratings import ColumnNames

from player_performance_ratings.predictor._base import BasePredictor
from player_performance_ratings.scorer.score import BaseScorer
from player_performance_ratings.transformation.base_transformer import BaseTransformer



class CrossValidator(ABC):

    def __init__(self, scorer: Optional[BaseScorer]):
        self.scorer = scorer


    @abstractmethod
    def generate_validation_df(self,
                               df: pd.DataFrame,
                               post_transformers: list[BaseTransformer],
                               column_names: ColumnNames,
                               predictor: BasePredictor,
                               estimator_features: list[str],
                               return_features: bool,
                               add_train_prediction: bool = False
                               ) -> pd.DataFrame:
        pass


    def cross_validation_score(self, validation_df: pd.DataFrame, scorer: Optional[BaseScorer] = None ) -> float:
        if not scorer and not self.scorer:
            raise ValueError("scorer is not defined. Either pass into constructor or as argument to method")

        scorer = scorer or self.scorer
        return scorer.score(df=validation_df)




import pandas as pd

df = pd.read_pickle("game_player.pickle")
df = df.drop_duplicates(subset=['game_id', 'player_id'])
gt = pd.read_pickle("game_team.pickle")
game = pd.read_pickle("game.pickle")
possessions = pd.read_pickle("possessions.pickle")
lineups = pd.read_pickle("lineups.pickle")

possessions = possessions.merge(lineups.rename(columns={'lineup': 'lineup_offense'}), left_on=['lineup_id_offense'] ,right_on=['lineup_id'], how='left')
possessions = possessions.merge(lineups.rename(columns={'lineup': 'lineup_defense'}), left_on=['lineup_id_defense'] ,right_on=['lineup_id'], how='left')

exploded = possessions.explode(['lineup_offense']).rename(columns={'lineup_offense': 'player_id'})
exploded['net_seconds'] = exploded['seconds_played_end'] - exploded['seconds_played_start']
grp = exploded.groupby(['game_id', 'player_id'])['points', 'net_seconds'].sum().reset_index()


q2 = possessions[possessions['game_id']=='0011900062']

game = game.rename(columns={'minutes': 'game_minutes'})
gt = gt.rename(columns={'pace': 'team_pace', 'possessions': 'team_possessions'})
df = df.merge(game[['game_id', 'game_minutes', 'start_date']], on='game_id', how='left')
df = df.merge(gt[['game_id', 'team_id', "team_id_opponent", 'score', 'score_opponent', 'won', "location",
                  "team_possessions", "team_pace"]],
              on=['game_id', 'team_id'], how='inner')
df.loc[df['won'] == True, 'won'] = 1
df.loc[df['won'] == False, 'won'] = 0

# randomly select 10% of game_ids

df = df[["team_id", "start_date", "game_id", "player_id", 'player_name', 'start_position', "team_id_opponent", "points",
         "game_minutes", 'pace', 'possessions',
         "minutes", "won", "plus_minus", "location", "score", "score_opponent", "team_possessions", "team_pace"]]
# df = df[df['start_date'].between('2022-10-17', '2023-02-01')]

df.to_pickle("game_player_full.pickle")



import copy
import logging
from dataclasses import dataclass, field
from typing import Optional, Union

import pandas as pd

from player_performance_ratings.transformation.pre_transformers import \
    SymmetricDistributionTransformer, MinMaxTransformer, PartialStandardScaler

from player_performance_ratings import ColumnNames

from player_performance_ratings.transformation.base_transformer import BaseTransformer


@dataclass
class ColumnWeight:
    name: str
    weight: float
    lower_is_better: bool = False

    def __post_init__(self):
        if self.weight < 0:
            raise ValueError("Weight must be positive")
        if self.weight > 1:
            raise ValueError("Weight must be less than 1")

@dataclass
class Performance:
    name: str = 'performance'
    weights: list[ColumnWeight] = field(default_factory=list)



def auto_create_pre_performance_transformations(
        pre_transformations: list[BaseTransformer],
        performances: list[Performance],
) -> list[BaseTransformer]:
    """
    Creates a list of transformers that ensure the performance column is generated in a way that makes sense for the rating model.
    Ensures columns aren't too skewed, scales them to similar ranges, ensure values are between 0 and 1,
    and then weights them according to the column_weights.
    """

    not_transformed_features = []
    transformed_features = []

    for performance in performances:
        not_transformed_features += [p.name for p in performance.weights if p.name not in not_transformed_features]

    distribution_transformer = SymmetricDistributionTransformer(features=not_transformed_features, prefix="")
    pre_transformations.append(distribution_transformer)

    all_features = transformed_features + not_transformed_features

    pre_transformations.append(
        PartialStandardScaler(features=all_features, ratio=1, max_value=9999, target_mean=0, prefix=""))
    pre_transformations.append(MinMaxTransformer(features=all_features))

    return pre_transformations


class PerformancesGenerator():

    def __init__(self,
                 performances: Union[list[Performance], Performance],
                 pre_transformations: Optional[list[BaseTransformer]] = None,
                 auto_transform_performance: bool = True
                 ):

        self.performances = performances if isinstance(performances, list) else [performances]
        self.auto_transform_performance = auto_transform_performance
        self.original_pre_transformations = [copy.deepcopy(p) for p in
                                             pre_transformations] if pre_transformations else []
        self.pre_transformations = pre_transformations or []
        if self.auto_transform_performance:
            self.pre_transformations = auto_create_pre_performance_transformations(
                pre_transformations=self.pre_transformations, performances=self.performances)


    def generate(self, df):


        if self.pre_transformations:
            for pre_transformation in self.pre_transformations:
                df = pre_transformation.fit_transform(df)

        for performance in self.performances:
            if self.pre_transformations:
                max_idx = len(self.pre_transformations) - 1
                column_weighs_mapping = {col_weight.name: self.pre_transformations[max_idx].features_out[idx] for
                                         idx, col_weight in enumerate(performance.weights)}
            else:
                column_weighs_mapping = None

            df[performance.name] = self._weight_columns(df=df,
                                                        performance_column_name=performance.name,
                                                        col_weights=performance.weights,
                                                        column_weighs_mapping=column_weighs_mapping)

            if df[performance.name].isnull().any():
                logging.error(
                    f"df[{performance.name}] contains nan values. Make sure all column_names used in column_weights are imputed beforehand")
                raise ValueError("performance contains nan values")

        return df

    def _weight_columns(self,
                        df: pd.DataFrame,
                        performance_column_name: str,
                        col_weights: list[ColumnWeight],
                        column_weighs_mapping: dict[str, str]
                        ) -> pd.DataFrame:
        df = df.copy()
        df[f"__{performance_column_name}"] = 0

        df['sum_cols_weights'] = 0
        for column_weight in col_weights:
            df[f'weight__{column_weight.name}'] = column_weight.weight
            df.loc[df[column_weight.name].isna(), f'weight__{column_weight.name}'] = 0
            df.loc[df[column_weight.name].isna(), column_weight.name] = 0
            df['sum_cols_weights'] = df['sum_cols_weights'] + df[f'weight__{column_weight.name}']

        drop_cols = ['sum_cols_weights', f"__{performance_column_name}"]
        for column_weight in col_weights:
            df[f'weight__{column_weight.name}'] / df['sum_cols_weights']
            drop_cols.append(f'weight__{column_weight.name}')

        sum_weight = sum([w.weight for w in col_weights])

        for column_weight in col_weights:

            if column_weighs_mapping:
                feature_name = column_weighs_mapping[column_weight.name]
            else:
                feature_name = column_weight.name

            if column_weight.lower_is_better:
                df[f"__{performance_column_name}"] += df[f'weight__{column_weight.name}'] / sum_weight * (
                        1 - df[feature_name])
            else:
                df[f"__{performance_column_name}"] += df[f'weight__{column_weight.name}'] / sum_weight * df[feature_name]

        return df[f"__{performance_column_name}"].clip(0, 1)

    @property
    def features_out(self) -> list[str]:
        return [c.name for c in self.performances]



import copy
import logging
from typing import Optional, Union

import pandas as pd

from player_performance_ratings.ratings.performances_generator import auto_create_pre_performance_transformations, \
    Performance
from player_performance_ratings.tuner.utils import get_default_lgbm_classifier_search_range

from player_performance_ratings.cross_validator.cross_validator import CrossValidator

from player_performance_ratings.pipeline import Pipeline
from player_performance_ratings.ratings import PerformancesGenerator, ColumnWeight
from player_performance_ratings.ratings.match_generator import convert_df_to_matches

from player_performance_ratings import PipelineFactory
from player_performance_ratings.tuner.predictor_tuner import PredictorTuner
from player_performance_ratings.tuner.rating_generator_tuner import RatingGeneratorTuner
from player_performance_ratings.tuner.performances_generator_tuner import PerformancesGeneratorTuner

logging.basicConfig(level=logging.INFO)


class PipelineTuner():
    """
    Given a search range, it will identify the best combination of pre_transformers, rating_generators and post_transformers.
    Using optuna, it will tune the hyperparameters of each transformer.
    """

    def __init__(self,
                 pipeline: Pipeline,
                 cross_validator: Optional[CrossValidator] = None,
                 performances_generator_tuners: Optional[
                     Union[list[Optional[PerformancesGeneratorTuner]], PerformancesGeneratorTuner]] = None,
                 rating_generator_tuners: Optional[
                     Union[list[Optional[RatingGeneratorTuner]], RatingGeneratorTuner]] = None,
                 predictor_tuner: Optional[PredictorTuner] = None,
                 fit_best: bool = True,
                 ):

        """
        :param scorer: The scorer to use to evaluate the performance of the match_predictor
        :param pipeline:
            The factory that creates the MatchPredictor.
            Contains the parameters to create the MatchPredictor if no parameter-tuning was done.
            Based on hyperparameter-tuning the final way the match-predictor is generated will be based on a combination of the parameters in the factory and the tuned parameters.
        :param performances_generator_tuners:
            The tuner that tunes the hyperparameters of the pre_transformers.
            If left none or as [], the pre_transformers in the pipeline_factory will be used.
        :param rating_generator_tuners:
            The tuner that tunes the hyperparameters of the rating_generators.
            If left none or as [], the rating_generators in the pipeline_factory will be used.

        :param predictor_tuner:
            The tuner that tunes the hyperparameters of the predictor model.

        :param fit_best: Whether to refit the match_predictor with the entire training use data using best hyperparameters
        """

        self.pipeline = pipeline
        self._pipeline_factory = PipelineFactory(
            rating_generators=pipeline.rating_generators,
            performances_generator=pipeline.performances_generator,
            post_rating_transformers=pipeline.post_rating_transformers,
            predictor=pipeline.predictor,
            column_names=pipeline.column_names
        )

        self.performances_generator_tuners = performances_generator_tuners or []
        if isinstance(self.performances_generator_tuners, PerformancesGeneratorTuner):
            self.performances_generator_tuners = [self.performances_generator_tuners]
        self.rating_generator_tuners = rating_generator_tuners or []
        self._untrained_best_model = None
        self.predictor_tuner = predictor_tuner
        if isinstance(self.rating_generator_tuners, RatingGeneratorTuner):
            self.rating_generator_tuners = [self.rating_generator_tuners]
        self.fit_best = fit_best

        if len(self.rating_generator_tuners) != len(
                self._pipeline_factory.rating_generators) and self.rating_generator_tuners:
            raise ValueError("Number of rating_generator_tuners must match number of rating_generators")

        if not self.pipeline.rating_generators and not self.pipeline.performances_generator and not self.predictor_tuner:

            if "LGBM" in self.pipeline.predictor.deepest_estimator.__class__.__name__:
                self.predictor_tuner = PredictorTuner(search_ranges=get_default_lgbm_classifier_search_range())
                logging.info("No tuning has been provided in config, using default LGBM search range")
            else:
                raise ValueError("No tuning has been provided in config")

        if self.performances_generator_tuners:
            performances = []
            for idx, tuner in enumerate(self.performances_generator_tuners):
                column_weights = []
                for parameter_search_range in tuner.performances_search_ranges[idx].search_ranges:
                    column_weights.append(
                        ColumnWeight(name=parameter_search_range.name, weight=0,
                                     lower_is_better=parameter_search_range.lower_is_better))
                performances.append(Performance(
                    name=self._pipeline_factory.rating_generators[idx].performance_column,
                    weights=column_weights
                ))

            if self._pipeline_factory.performances_generator:
                original_pre_transformations = self._pipeline_factory.performances_generator.original_pre_transformations
                other_params = {k: v for k, v in self._pipeline_factory.performances_generator.__dict__.items() if
                                k not in ["performances", "pre_transformations", "original_pre_transformations"]}

            else:
                original_pre_transformations = []
                other_params = {}

            self._pipeline_factory.performances_generator = PerformancesGenerator(performances=performances,
                                                                                  pre_transformations=original_pre_transformations,
                                                                                  **other_params)

        self.cross_validator = cross_validator

    def tune(self,
             df: pd.DataFrame,
             return_df: bool = False,
             return_cross_validated_predictions: bool = False,
             ) -> Union[Pipeline, tuple[Pipeline, pd.DataFrame]]:

        if self.cross_validator is None:
            self.cross_validator = self.pipeline.create_default_cross_validator(df)

        original_df = df.copy()

        best_performances_generator: PerformancesGenerator = copy.deepcopy(
            self._pipeline_factory.performances_generator)

        untrained_best_performances_generator = copy.deepcopy(best_performances_generator)

        if self.performances_generator_tuners:
            for rating_idx, performances_generator_tuner in enumerate(self.performances_generator_tuners):
                if performances_generator_tuner is None:
                    continue
                logging.info("Tuning PreTransformers")
                best_performances_generator = performances_generator_tuner.tune(df=df,
                                                                                rating_idx=rating_idx,
                                                                                pipeline_factory=self._pipeline_factory,
                                                                                cross_validator=self.cross_validator)
            untrained_best_performances_generator = copy.deepcopy(best_performances_generator)
        if best_performances_generator:
            df = best_performances_generator.generate(df)

        best_rating_generators = copy.deepcopy(self._pipeline_factory.rating_generators)

        matches = []
        for rating_generator in self.pipeline.rating_generators:
            rating_matches = convert_df_to_matches(df=df, column_names=self.pipeline.column_names,
                                                   performance_column_name=rating_generator.performance_column)
            matches.append(rating_matches)

        rating_generators = self._pipeline_factory.rating_generators
        untrained_best_rating_generators = copy.deepcopy(best_rating_generators)
        for rating_idx, rating_generator in enumerate(rating_generators):

            if self.rating_generator_tuners:
                rating_generator_tuner = self.rating_generator_tuners[rating_idx]
                if rating_generator_tuner is None:
                    continue

                rating_matches = convert_df_to_matches(df=df, column_names=self.pipeline.column_names,
                                                       performance_column_name=rating_generator.performance_column)
                matches.append(rating_matches)
                tuned_rating_generator = rating_generator_tuner.tune(df=df, matches=matches[rating_idx],
                                                                     rating_idx=rating_idx,
                                                                     cross_validator=self.cross_validator,
                                                                     pipeline_factory=self._pipeline_factory)

                best_rating_generators[rating_idx] = tuned_rating_generator

                untrained_best_rating_generators[rating_idx] = copy.deepcopy(tuned_rating_generator)
            match_ratings = best_rating_generators[rating_idx].generate_historical(df=df, matches=matches[rating_idx],
                                                                                   column_names=self.pipeline.column_names)

            for rating_feature in best_rating_generators[rating_idx].estimator_features_return:
                values = match_ratings[rating_feature]

                if len(self.rating_generator_tuners) > 1:
                    rating_feature_str = rating_feature + str(rating_idx)
                else:
                    rating_feature_str = rating_feature

                df = df.assign(**{rating_feature_str: values})

        best_post_transformers = copy.deepcopy(self._pipeline_factory.post_rating_transformers)
        untrained_best_post_transformers = copy.deepcopy(best_post_transformers)

        if self.predictor_tuner:
            logging.info("Tuning Predictor")
            best_predictor = self.predictor_tuner.tune(df=df, cross_validator=self.cross_validator,
                                                       pipeline_factory=self._pipeline_factory,
                                                       best_post_rating_transformers=untrained_best_post_transformers)
            untrained_best_predictor = copy.deepcopy(best_predictor)
        else:
            untrained_best_predictor = copy.deepcopy(self._pipeline_factory.predictor)

        self._untrained_best_model = Pipeline(
            rating_generators=[copy.deepcopy(rating_generator) for rating_generator in
                               untrained_best_rating_generators],
            performances_generator=copy.deepcopy(untrained_best_performances_generator),
            post_rating_transformers=[copy.deepcopy(post_rating_transformer) for post_rating_transformer in
                                      untrained_best_post_transformers],
            predictor=untrained_best_predictor,
            column_names=self.pipeline.column_names
        )
        best_match_predictor = Pipeline(
            rating_generators=[copy.deepcopy(rating_generator) for rating_generator in
                               untrained_best_rating_generators],
            performances_generator=copy.deepcopy(untrained_best_performances_generator),
            post_rating_transformers=[copy.deepcopy(post_rating_transformer) for post_rating_transformer in
                                      untrained_best_post_transformers],
            predictor=untrained_best_predictor,
            column_names=self.pipeline.column_names
        )

        if self.fit_best:
            logging.info("Retraining best match predictor with all data")
            df = best_match_predictor.train_predict(df=original_df, store_ratings=True, return_features=True,
                                                    cross_validate_predict=return_cross_validated_predictions,
                                                    cross_validator=self.cross_validator)

        if return_df:
            return best_match_predictor, df
        return best_match_predictor

    @property
    def untrained_best_model(self):
        return self._untrained_best_model



import logging
from abc import abstractmethod, ABC
from typing import Optional

import pandas as pd
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import OneHotEncoder, StandardScaler

from player_performance_ratings.predictor_transformer import PredictorTransformer, SkLearnTransformerWrapper, \
    ConvertDataFrameToCategoricalTransformer



class BasePredictor(ABC):

    def __init__(self,
                 estimator,
                 estimator_features: list[str],
                 target: str,
                 pre_transformers: Optional[list[PredictorTransformer]] = None,
                 pred_column: Optional[str] = None,
                 filters: Optional[dict] = None
                 ):
        self._estimator_features = estimator_features or []
        self.estimator = estimator
        self._target = target
        self._pred_column = pred_column or f"{self._target}_prediction"
        self.pre_transformers = pre_transformers or []
        self._deepest_estimator = self.estimator
        self.filters = filters or []
        self.multiclassifier = False

        iterations = 0
        while hasattr(self._deepest_estimator, "estimator"):
            self._deepest_estimator = self._deepest_estimator.estimator
            iterations += 1
            if iterations > 10:
                raise ValueError("estimator is too deep")

    @property
    def estimator_type(self) -> str:
        if hasattr(self._deepest_estimator, "predict_proba"):
            return "classifier"
        return "regressor"

    @property
    def deepest_estimator(self) -> object:
        return self._deepest_estimator

    @abstractmethod
    def train(self, df: pd.DataFrame, estimator_features: list[str]) -> None:
        pass

    @abstractmethod
    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @property
    def pred_column(self) -> str:
        return self._pred_column

    @property
    def target(self) -> str:
        return self._target

    @property
    def columns_added(self) -> list[str]:
        if not self.multiclassifier:
            return [self.pred_column]
        return [self.pred_column, "classes"]

    @property
    def classes_(self) -> Optional[list[str]]:
        if 'classes_' not in dir(self.estimator):
            return None
        return self.estimator.classes_

    def set_target(self, new_target_name: str):
        self._target = new_target_name

    def fit_transform_pre_transformers(self, df: pd.DataFrame) -> pd.DataFrame:

        feats_to_transform = []
        for estimator_feature in self._estimator_features.copy():

            if estimator_feature not in df.columns:
                self._estimator_features.remove(estimator_feature)

            elif df[estimator_feature].dtype in ('str', 'object') and estimator_feature not in [f.features[0] for f in
                                                                                                self.pre_transformers]:
                feats_to_transform.append(estimator_feature)

        if feats_to_transform:
            if self._deepest_estimator.__class__.__name__ in ('LogisticRegression', 'LinearRegression'):
                logging.info(f"Adding OneHotEncoder to pre_transformers for features: {feats_to_transform}")
                self.pre_transformers.append(
                    SkLearnTransformerWrapper(transformer=OneHotEncoder(handle_unknown='ignore'),
                                              features=feats_to_transform))

            elif self._deepest_estimator.__class__.__name__ in ('LGBMRegressor', 'LGBMClassifier'):
                logging.info(
                    f"Adding ConvertDataFrameToCategoricalTransformer to pre_transformers for features: {feats_to_transform}")
                self.pre_transformers.append(ConvertDataFrameToCategoricalTransformer(features=feats_to_transform))

        for pre_transformer in self.pre_transformers:
            for feature in pre_transformer.features:
                if feature in self._estimator_features:
                    self._estimator_features.remove(feature)


        if self._deepest_estimator.__class__.__name__ in (
        'LogisticRegression', 'LinearRegression'):
            if 'StandardScaler' not in [
                pre_transformer.transformer.__class__.__name__ for pre_transformer in
                self.pre_transformers if hasattr(pre_transformer, "transformer")]:
                logging.info(f"Adding StandardScaler to pre_transformers")
                self.pre_transformers.append(SkLearnTransformerWrapper(transformer=StandardScaler(),
                                                                       features=self._estimator_features.copy()))

            if 'SimpleImputer' not in [
                pre_transformer.transformer.__class__.__name__ for pre_transformer in
                self.pre_transformers if hasattr(pre_transformer, "transformer")]:
                self.pre_transformers.append(SkLearnTransformerWrapper(transformer=SimpleImputer(),
                                                                       features=self._estimator_features.copy()))

        for pre_transformer in self.pre_transformers:
            df = pre_transformer.fit_transform(df)
            self._estimator_features = list(set(pre_transformer.features_out + self._estimator_features))

        return df

    def transform_pre_transformers(self, df: pd.DataFrame) -> pd.DataFrame:
        if self.pre_transformers:
            for pre_transformer in self.pre_transformers:
                df = pre_transformer.transform(df)
        return df

    @property
    def estimator_features(self) -> list[str]:
        return self._estimator_features



import copy
import inspect
import logging
from dataclasses import dataclass

import pandas as pd
from player_performance_ratings.data_structures import Match, ColumnNames

from player_performance_ratings.cross_validator._base import CrossValidator

from player_performance_ratings import PipelineFactory
from player_performance_ratings.ratings import UpdateRatingGenerator
from player_performance_ratings.ratings.enums import RatingEstimatorFeatures, RatingHistoricalFeatures
from player_performance_ratings.ratings.rating_calculators.start_rating_generator import \
    StartRatingGenerator


@dataclass
class LeagueH2H:
    league: str
    opp_league: str
    weight: float
    mean_rating_change: float


class StartLeagueRatingOptimizer():

    def __init__(self,
                 pipeline_factory: PipelineFactory,
                 cross_validator: CrossValidator,
                 max_iterations: int = 2,
                 learning_step: int = 20,
                 weight_div: int = 500,
                 indirect_weight: float = 1.5,
                 verbose: int = 1,
                 ):
        self.pipeline_factory = pipeline_factory
        self.max_iterations = max_iterations
        self.learning_step = learning_step
        self.weight_div = weight_div
        self.cross_validator = cross_validator
        self.indirect_weight = indirect_weight
        self.verbose = verbose
        self._scores = []
        self._league_ratings_iterations = []

    def optimize(self,
                 df: pd.DataFrame,
                 rating_model_idx: int,
                 rating_generator: UpdateRatingGenerator,
                 matches: list[Match],
                 column_names: ColumnNames,
                 ) -> dict[str, float]:



        start_rating_generator = rating_generator.match_rating_generator.start_rating_generator
        league_ratings = start_rating_generator.league_ratings.copy()
        start_rating_params = list(
            inspect.signature(start_rating_generator.__class__.__init__).parameters.keys())[1:]
        non_league_ratings_params = {attr: getattr(start_rating_generator, attr) for attr in
                                     dir(start_rating_generator) if
                                     attr in start_rating_params and attr != 'league_ratings'}

        for iteration in range(self.max_iterations + 1):

            rating_generator_used = copy.deepcopy(rating_generator)

            start_rating_generator = StartRatingGenerator(league_ratings=league_ratings, **non_league_ratings_params)

            if iteration == self.max_iterations:
                min_idx = self._scores.index(min(self._scores))
                logging.info(f"best start rating params {self._league_ratings_iterations[min_idx]}")
                return self._league_ratings_iterations[min_idx]

            rating_generator_used.match_rating_generator.start_rating_generator = start_rating_generator

            rating_generators = copy.deepcopy(self.pipeline_factory.rating_generators)
            rating_generators[rating_model_idx] = rating_generator_used

            rating_values = rating_generators[rating_model_idx].generate_historical(df=df, matches=matches, column_names=column_names)
            for rating_column, rating_value in rating_values.items():
                df = df.assign(**{rating_column: rating_value})

            ratings_df = rating_generators[rating_model_idx].ratings_df
            df_incl_ratings = df.merge(ratings_df[[column_names.match_id, column_names.team_id, column_names.player_id,
                                                   RatingHistoricalFeatures.PLAYER_RATING_CHANGE,
                                                   RatingEstimatorFeatures.PLAYER_LEAGUE,
                                                   RatingEstimatorFeatures.OPPONENT_LEAGUE,
                                                   ]],
                                       on=[column_names.match_id, column_names.player_id, column_names.team_id],
                                       how='inner')

            rating_generators = copy.deepcopy(self.pipeline_factory.rating_generators)
            rating_generators[rating_model_idx].match_rating_generator.start_rating_generator = start_rating_generator
            pipeline = self.pipeline_factory.create(rating_generators=rating_generators)
            score = pipeline.cross_validate_score(df=df_incl_ratings, cross_validator=self.cross_validator,
                                                  create_performance=False, create_rating_features=False)

            if not league_ratings:
                league_ratings = pipeline.rating_generators[
                    rating_model_idx].match_rating_generator.start_rating_generator.league_ratings

            self._league_ratings_iterations.append(copy.deepcopy(league_ratings))
            self._scores.append(score)
            if self.verbose:
                logging.info(f"iteration {iteration} finished. Score: {score}. best startings {league_ratings}")

            league_rating_changes = (df_incl_ratings
            .groupby([RatingEstimatorFeatures.PLAYER_LEAGUE, RatingEstimatorFeatures.OPPONENT_LEAGUE])
            .agg(
                {
                    RatingHistoricalFeatures.PLAYER_RATING_CHANGE: 'mean',
                    column_names.player_id: 'count'
                }
            )
            .reset_index()
            .rename(
                columns={
                    RatingHistoricalFeatures.PLAYER_RATING_CHANGE: 'mean_rating_change',
                    column_names.player_id: 'count'
                })
            )
            leagues = league_rating_changes[RatingEstimatorFeatures.PLAYER_LEAGUE].unique().tolist()

            league_opp_league_h2hs: dict[str, dict[str, LeagueH2H]] = {}
            league_to_played_against_leagues = {}

            for league in leagues:
                league_to_played_against_leagues[league] = \
                    league_rating_changes[league_rating_changes[RatingEstimatorFeatures.PLAYER_LEAGUE] == league][
                        RatingEstimatorFeatures.OPPONENT_LEAGUE].unique().tolist()

                league_opp_league_h2hs[league] = {}
                for opp_league in leagues:
                    if league == opp_league:
                        continue

                    rows = league_rating_changes[(league_rating_changes[RatingEstimatorFeatures.PLAYER_LEAGUE] == league) &
                                                 (league_rating_changes[
                                                      RatingEstimatorFeatures.OPPONENT_LEAGUE] == opp_league)]
                    if len(rows) == 0:
                        weight = 0
                        mean_rating_change = 0
                    else:
                        count = rows['count'].iloc[0]
                        weight = min(1, count / self.weight_div)
                        mean_rating_change = rows['mean_rating_change'].iloc[0]

                    league_h2h = LeagueH2H(
                        weight=weight,
                        mean_rating_change=mean_rating_change,
                        league=league,
                        opp_league=opp_league,
                    )
                    league_opp_league_h2hs[league][opp_league] = league_h2h

            final_league_opp_league_h2hs: dict[str, dict[str, LeagueH2H]] = {}

            league_sum_final_weights = {}
            for league, opp_league_h2h in league_opp_league_h2hs.items():

                league_sum_final_weights[league] = 0
                final_league_opp_league_h2hs[league] = {}
                for opp_league, h2h in opp_league_h2h.items():

                    if h2h.weight < 1:
                        league_against = league_to_played_against_leagues[league]
                        opp_played_against = league_to_played_against_leagues[opp_league]
                        shared_leagues = [l for l in league_against if
                                          l in opp_played_against and l not in (league, opp_league)]

                        relative_shared_weighted_mean = 0
                        shared_weight = 0
                        for shared_league in shared_leagues:
                            opp_vs_shared_h2h = league_opp_league_h2hs[opp_league][shared_league]
                            league_vs_shared_h2h = league_opp_league_h2hs[league][shared_league]

                            sum_rating_change = league_vs_shared_h2h.mean_rating_change - opp_vs_shared_h2h.mean_rating_change

                            relative_shared_weighted_mean += min(
                                opp_vs_shared_h2h.weight, league_vs_shared_h2h.weight) * sum_rating_change
                            shared_weight += min(
                                opp_vs_shared_h2h.weight, league_vs_shared_h2h.weight)

                    weighted_mean = h2h.mean_rating_change * h2h.weight + min(1,
                                                                              shared_weight * self.indirect_weight) * relative_shared_weighted_mean
                    final_weight = min(1, h2h.weight + shared_weight)
                    final_league_h2h = LeagueH2H(
                        weight=final_weight,
                        mean_rating_change=weighted_mean,
                        league=league,
                        opp_league=opp_league,
                    )
                    final_league_opp_league_h2hs[league][opp_league] = final_league_h2h
                    league_sum_final_weights[league] += final_weight

            for league, final_opp_league_h2h in final_league_opp_league_h2hs.items():

                if league not in league_ratings:
                    continue
                start_rating_rating_change = 0

                for opp_league, final_h2h in final_opp_league_h2h.items():
                    if league_sum_final_weights[
                        league] > 0:
                        start_rating_rating_change += final_h2h.weight / league_sum_final_weights[
                            league] * final_h2h.mean_rating_change * self.learning_step

                league_ratings[league] += start_rating_rating_change

        return league_ratings



import logging
from dataclasses import dataclass, field
from typing import Literal, Optional, Union, Any

from optuna.trial import BaseTrial


@dataclass
class ParameterSearchRange:
    name: str
    type: Literal["uniform", "loguniform", "int", "categorical", "discrete_uniform"]
    low: Optional[Union[float, int]] = None
    high: Optional[Union[float, int]] = None
    choices: Optional[list[Any]] = None
    lower_is_better: bool = False
    custom_params: Optional[dict[str, Any]] = None,

    def __post_init__(self):
        if self.custom_params is None:
            self.custom_params = {}


def add_params_from_search_range(trial: BaseTrial, parameter_search_range: list[ParameterSearchRange],
                                 params: dict) -> dict:
    for config in parameter_search_range:
        if config.type == "uniform":
            params[config.name] = trial.suggest_uniform(config.name, low=config.low, high=config.high)
        elif config.type == "loguniform":
            params[config.name] = trial.suggest_loguniform(config.name, low=config.low, high=config.high)
        elif config.type == "int":
            params[config.name] = trial.suggest_int(config.name, low=config.low, high=config.high)
        elif config.type == "categorical":
            params[config.name] = trial.suggest_categorical(config.name, config.choices)
        else:
            logging.warning(f"Unknown type {config.type} for parameter {config.name}")

    return params


def get_default_lgbm_classifier_search_range() -> list[ParameterSearchRange]:
    return [
        ParameterSearchRange(
            name='learning_rate',
            type='uniform',
            low=0.02,
            high=0.1,
        ),
        ParameterSearchRange(
            name='n_estimators',
            type='int',
            low=40,
            high=800,
        ),
        ParameterSearchRange(
            name='num_leaves',
            type='int',
            low=10,
            high=100,
        ),
        ParameterSearchRange(
            name='max_depth',
            type='int',
            low=2,
            high=10,
        ),
        ParameterSearchRange(
            name='min_child_samples',
            type='int',
            low=2,
            high=200,
        ),
        ParameterSearchRange(
            name='reg_alpha',
            type='uniform',
            low=0,
            high=5,
        ),
    ]


def get_default_lgbm_regressor_search_range() -> list[ParameterSearchRange]:

    return [
        ParameterSearchRange(
            name='learning_rate',
            type='uniform',
            low=0.02,
            high=0.1,
        ),
        ParameterSearchRange(
            name='n_estimators',
            type='int',
            low=50,
            high=1000,
        ),
        ParameterSearchRange(
            name='num_leaves',
            type='int',
            low=10,
            high=100,
        ),
        ParameterSearchRange(
            name='max_depth',
            type='int',
            low=2,
            high=14,
        ),
        ParameterSearchRange(
            name='min_child_samples',
            type='int',
            low=2,
            high=200,
        ),
        ParameterSearchRange(
            name='reg_alpha',
            type='uniform',
            low=0,
            high=5,
        ),
    ]


def get_default_team_rating_search_range() -> list[ParameterSearchRange]:
    return [
        ParameterSearchRange(
            name='confidence_weight',
            type='uniform',
            low=0.7,
            high=0.95
        ),
        ParameterSearchRange(
            name='confidence_days_ago_multiplier',
            type='uniform',
            low=0.02,
            high=.12,
        ),
        ParameterSearchRange(
            name='confidence_max_days',
            type='uniform',
            low=60,
            high=220,
        ),
        ParameterSearchRange(
            name='confidence_max_sum',
            type='uniform',
            low=60,
            high=300,
        ),
        ParameterSearchRange(
            name='confidence_value_denom',
            type='uniform',
            low=25,
            high=120
        ),
        ParameterSearchRange(
            name='rating_change_multiplier',
            type='uniform',
            low=10,
            high=80
        ),
        ParameterSearchRange(
            name='min_rating_change_multiplier_ratio',
            type='uniform',
            low=0.02,
            high=0.2,
        ),
        ParameterSearchRange(
            name='team_id_change_confidence_sum_decrease',
            type='uniform',
            low=0,
            high=15,
        )
    ]



from .enums import RatingEstimatorFeatures, RatingHistoricalFeatures, PredictedRatingMethod, InputColumnNames
from .league_identifier import LeagueIdentifier
from .match_generator import convert_df_to_matches
from .time_weight_ratings import BayesianTimeWeightedRating
from .performances_generator import ColumnWeight, PerformancesGenerator
from .update_rating_generator import UpdateRatingGenerator
from .rating_calculators.match_rating_generator import MatchRatingGenerator
from .rating_calculators.start_rating_generator import StartRatingGenerator
from .rating_calculators.performance_predictor import RatingDifferencePerformancePredictor, \
    RatingMeanPerformancePredictor, RatingNonOpponentPerformancePredictor



import math
from typing import Optional, Union

import numpy as np
import pandas as pd

from player_performance_ratings.data_structures import Match, ColumnNames, TeamRating, PlayerRating
from player_performance_ratings.ratings import convert_df_to_matches
from player_performance_ratings.ratings.enums import RatingEstimatorFeatures, RatingHistoricalFeatures
from player_performance_ratings.ratings.rating_generator import RatingGenerator


# NOT WORKING AT THE MOMENT

class BayesianTimeWeightedRating(RatingGenerator):
    """
    Generates ratings for players and teams by watching past performances higher using a combination of prior, evidence and likelihood.
    """

    def __init__(self,
                 column_names: ColumnNames,
                 evidence_exponential_weight: float = 0.96,
                 likelihood_exponential_weight: float = 0.98,
                 likelihood_denom: float = 50,
                 prior_granularity_count_max: int = 200,
                 prior_by_league: bool = True,
                 prior_by_position: bool = True,
                 features_created: Optional[list[str]] = None,
                 ):

        """

        :param evidence_exponential_weight:
            The weight of the evidence. Value should be between 0 and 1.
            A value closer to 0 makes performances further in the past less important.

        :param likelihood_exponential_weight:
            Determines how the likelihood_ratio is calculated. The likelihood ratio determines how much the prior and evidence are weighted.
            A likelihood_exponential_weight closer to 0 makes the prior have a larger weight relative to the evidence.

        :param likelihood_denom:
            The denominator of the likelihood_ratio.
            An increase in the likelihood_denom makes the player require more data (and with higher recency) for the likelihood_ratio to approach 1.
            Thus a higher likelihood_denom makes the prior have a larger weight relative to the evidence.

        :param prior_granularity_count_max:
            In calculating the prior-rating, past ratings of players in similar groups/granularity are used.
            If no players exist in the group at the time the prior of a player is detemrined, the player will receive the mean of all performances as his rating
            The more players exist in the group, the more the prior-rating will be based on the past performances of players in the group.
            A higher prior_granularity_count_max results in a lower weight of the within-group-players-rating relative to the mean of all performances across all players in the dataset.


        :param prior_by_league:
            If True, the prior-rating will be based on ratings of players within the same league
        :param prior_by_position:
            If True, the prior-rating will be based on ratings of players within the same position
        """

        super().__init__(column_names=column_names)
        if evidence_exponential_weight < 0 or evidence_exponential_weight > 1:
            raise ValueError("evidence_exponential_weight must be between 0 and 1")
        if likelihood_exponential_weight < 0 or likelihood_exponential_weight > 1:
            raise ValueError("likelihood_exponential_weight must be between 0 and 1")

        self._player_ratings: dict[str, PlayerRating] = {}
        self.evidence_exponential_weight = evidence_exponential_weight
        self.likelihood_exponential_weight = likelihood_exponential_weight
        self.likelihood_denom = likelihood_denom
        self.prior_granularity_count_max = prior_granularity_count_max
        self.by_league = prior_by_league
        self.by_position = prior_by_position
        self._features_out = features_created or [RatingEstimatorFeatures.TIME_WEIGHTED_RATING,
                                                  RatingEstimatorFeatures.TIME_WEIGHTED_RATING_EVIDENCE,
                                                  RatingEstimatorFeatures.TIME_WEIGHTED_RATING_LIKELIHOOD_RATIO]

        self.player_performances: dict[str, list[float]] = {}
        self.player_days: dict[str, list[int]] = {}
        self._player_prior_ratings: dict[str, float] = {}
        self._granularity_ratings: dict[str, list[float]] = {}
        self._granularity_players: dict[str, list[str]] = {}

    def generate_historical(self, df: Optional[pd.DataFrame] = None, matches: Optional[list[Match]] = None) -> dict[RatingEstimatorFeatures, list[float]]:

        if matches is None and df is None:
            raise ValueError("If matches is not passed, df and column names must be massed")

        if matches is None:
            matches = convert_df_to_matches(df=df, column_names=self.column_names)

        if df is not None:
            mean_performance_value = df[self.column_names.performance].mean()
        else:
            count = 0
            sum_mean_performance_value = 0
            for match in matches:
                for team in match.teams:
                    for team_player in team.players:
                        count += 1
                        sum_mean_performance_value += team_player.performance.performance_value

            mean_performance_value = sum_mean_performance_value / count

        ratings = {
            RatingEstimatorFeatures.TIME_WEIGHTED_RATING: [],
            RatingEstimatorFeatures.TIME_WEIGHTED_RATING_LIKELIHOOD_RATIO: [],
            RatingEstimatorFeatures.TIME_WEIGHTED_RATING_EVIDENCE: [],
        }

        for match in matches:
            for team in match.teams:
                for team_player in team.players:
                    self.player_performances.setdefault(team_player.id, [])
                    self.player_days.setdefault(team_player.id, [])

                    days_agos = match.day_number - np.array(self.player_days[team_player.id])
                    weights = np.power(self.evidence_exponential_weight, days_agos)
                    evidence_performances = np.sum(
                        np.array(self.player_performances[team_player.id]) * weights) / np.sum(
                        weights)
                    if math.isnan(evidence_performances):
                        evidence_performances = None
                    likelihood_exponential_weights = np.power(self.likelihood_exponential_weight, days_agos)

                    prior_rating = self._generate_base_prior(league=team_player.league, position=team_player.position,
                                                             base_prior_value=mean_performance_value,
                                                             player_id=team_player.id)

                    likelihood_ratio = min(sum(likelihood_exponential_weights) / self.likelihood_denom, 1)
                    posterior_rating = (
                                               1 - likelihood_ratio) * prior_rating + likelihood_ratio * evidence_performances if evidence_performances is not None else prior_rating

                    self._update_granularity_ratings(league=team_player.league, player_rating=posterior_rating,
                                                     position=team_player.position, player_id=team_player.id)

                    self.player_days[team_player.id].append(match.day_number)
                    self.player_performances[team_player.id].append(team_player.performance.performance_value)

                    ratings[RatingEstimatorFeatures.TIME_WEIGHTED_RATING_LIKELIHOOD_RATIO].append(likelihood_ratio)
                    ratings[RatingEstimatorFeatures.TIME_WEIGHTED_RATING].append(posterior_rating)
                    ratings[RatingEstimatorFeatures.TIME_WEIGHTED_RATING_EVIDENCE].append(evidence_performances)

        return ratings

    def generate_future(self, matches: Optional[list[Match]] = None, df: Optional[pd.DataFrame] = None) -> dict[
        RatingEstimatorFeatures, list[float]]:
        raise NotImplementedError("This method is not implemented")

    def _generate_base_prior(self, player_id: str, base_prior_value: float, league: Optional[str],
                             position: Optional[str]) -> float:
        granularity_id = ""
        if self.by_league:
            league = league if league is not None else ""
            granularity_id += league
        if self.by_position:
            position = position if position is not None else ""
            granularity_id += position

        if player_id in self._player_prior_ratings and granularity_id in self._granularity_players and player_id in \
                self._granularity_players[granularity_id]:
            return self._player_prior_ratings[player_id]

        self._granularity_ratings.setdefault(granularity_id, [])
        self._granularity_players.setdefault(granularity_id, [])
        granularity_weight = len(self._granularity_ratings[granularity_id]) / self.prior_granularity_count_max

        if len(self._granularity_ratings[granularity_id]) == 0:
            self._player_prior_ratings[player_id] = base_prior_value
            return base_prior_value

        prior_rating = granularity_weight * np.mean(self._granularity_ratings[granularity_id]) + (
                1 - granularity_weight) * base_prior_value
        self._player_prior_ratings[player_id] = prior_rating
        return prior_rating

    def _update_granularity_ratings(self, player_id: str, player_rating: float, league: Optional[str],
                                    position: Optional[str]):
        granularity_id = ""
        if self.by_league:
            league = league if league is not None else ""
            granularity_id += league
        if self.by_position:
            position = position if position is not None else ""
            granularity_id += position

        if player_id in self._granularity_players[granularity_id]:
            index = self._granularity_players[granularity_id].index(player_id)
            self._granularity_ratings[granularity_id][index] = player_rating
            return

        self._granularity_players[granularity_id].append(player_id)
        self._granularity_ratings[granularity_id].append(player_rating)

    @property
    def player_ratings(self) -> dict[str, PlayerRating]:
        return self._player_ratings

    @property
    def team_ratings(self) -> list[TeamRating]:
        return self.team_ratings

    @property
    def estimator_features_out(self) -> list[RatingEstimatorFeatures]:
        return self._features_out

    @property
    def features_out(self) -> list[Union[RatingEstimatorFeatures, RatingHistoricalFeatures]]:
        return self._features_out



import mock
import pandas as pd

from player_performance_ratings.cross_validator import MatchKFoldCrossValidator, MatchCountCrossValidator
from player_performance_ratings.predictor import Predictor
from sklearn.linear_model import LinearRegression

from player_performance_ratings.ratings import BayesianTimeWeightedRating, \
    RatingEstimatorFeatures, convert_df_to_matches, ColumnWeight, UpdateRatingGenerator, PerformancesGenerator, \
    MatchRatingGenerator
from player_performance_ratings.ratings.performances_generator import Performance

from player_performance_ratings.transformation import LagTransformer

from player_performance_ratings import ColumnNames, Pipeline


def test_match_predictor_auto_pre_transformers():
    df = pd.DataFrame({
        "game_id": [1, 1, 2, 2, 3, 3],
        "player_id": [1, 2, 3, 1, 2, 3],
        "team_id": [1, 2, 1, 2, 1, 3],
        "start_date": [pd.to_datetime("2023-01-01"), pd.to_datetime("2023-01-01"), pd.to_datetime("2023-01-02"),
                       pd.to_datetime("2023-01-02"), pd.to_datetime("2023-01-03"), pd.to_datetime("2023-01-03")],
        'deaths': [1, 1, 1, 2, 2, 2],
        "kills": [0.2, 0.3, 0.4, 0.5, 2, 0.2],
        "__target": [1, 0, 1, 0, 1, 0],
    })

    column_weights = [
        ColumnWeight(name="kills", weight=0.6),
        ColumnWeight(name="deaths", weight=0.4, lower_is_better=True)
    ]

    expected_df = df.copy()
    expected_df["prediction"] = [0.5, 0.5, 0.5, 0.5, 0.5, 0.5]

    predictor_mock = mock.Mock()
    predictor_mock.target = "__target"
    predictor_mock.columns_added = ['prediction']
    predictor_mock._estimator_features = [RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED]

    predictor_mock.add_prediction.return_value = expected_df
    rating_generators = UpdateRatingGenerator(
        estimator_features_out=[RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED],
        performance_column="weighted_performance"
    )

    pipeline = Pipeline(
        predictor=predictor_mock,
        rating_generators=rating_generators,
        performances_generator=PerformancesGenerator(
            performances=Performance(name="weighted_performance", weights=column_weights)),
        column_names=ColumnNames(
            match_id="game_id",
            team_id="team_id",
            player_id="player_id",
            start_date="start_date",
        )
    )

    new_df = pipeline.train_predict(df=df)

    pd.testing.assert_frame_equal(new_df, expected_df, check_like=True)

    assert len(pipeline.performances_generator.pre_transformations) > 0


def test_match_predictor_multiple_rating_generators_same_performance():
    df = pd.DataFrame({
        "game_id": [1, 1, 2, 2, 3, 3],
        "player_id": [1, 2, 3, 1, 2, 3],
        "team_id": [1, 2, 1, 2, 1, 3],
        "start_date": [pd.to_datetime("2023-01-01"), pd.to_datetime("2023-01-01"), pd.to_datetime("2023-01-02"),
                       pd.to_datetime("2023-01-02"), pd.to_datetime("2023-01-03"), pd.to_datetime("2023-01-03")],
        "performance": [0.2, 0.8, 0.4, 0.6, 1, 0],
        "__target": [1, 0, 1, 0, 1, 0],
    })

    column_names1 = ColumnNames(
        match_id="game_id",
        team_id="team_id",
        player_id="player_id",
        start_date="start_date",
    )

    df = df.sort_values(
        by=[column_names1.start_date, column_names1.match_id,
            column_names1.team_id, column_names1.player_id])

    expected_df = df.copy()
    expected_df["prediction"] = [0.5, 0.5, 0.5, 0.5, 0.5, 0.5]

    predictor_mock = mock.Mock()
    predictor_mock.target = "__target"
    predictor_mock.add_prediction.return_value = expected_df
    predictor_mock.columns_added = ['prediction']
    predictor_mock._estimator_features = [RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED]

    match_predictor = Pipeline(
        rating_generators=[
            UpdateRatingGenerator(estimator_features_out=[RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED]),
            UpdateRatingGenerator(match_rating_generator=MatchRatingGenerator(rating_change_multiplier=20),
                                  estimator_features_out=[RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED])],
        post_rating_transformers=[],
        predictor=predictor_mock,
        column_names=column_names1
    )

    new_df = match_predictor.train_predict(df=df)
    pd.testing.assert_frame_equal(new_df, expected_df, check_like=True)

    col_names_predictor_train = predictor_mock.train.call_args[1]['df'].columns.tolist()

    col_names_predictor_add = predictor_mock.add_prediction.call_args[1]['df'].columns.tolist()

    assert RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED + str(1) in col_names_predictor_add
    assert RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED + str(1) in col_names_predictor_train

    assert RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED + str(0) in col_names_predictor_add
    assert RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED + str(0) in col_names_predictor_train


def test_match_predictor_0_rating_generators():
    """
    Post rating transformers are used, but no rating model. the features from transformers should be used to train model and add prediction
    """

    df = pd.DataFrame({
        "game_id": [1, 1, 2, 2, 3, 3],
        "player_id": [1, 2, 3, 1, 2, 3],
        "team_id": [1, 2, 1, 2, 1, 3],
        "start_date": [1, 1, 2, 2, 3, 3],
        'deaths': [1, 1, 1, 2, 2, 2],
        "kills": [0.2, 0.3, 0.4, 0.5, 2, 0.2],
        "__target": [1, 0, 1, 0, 1, 0],
    })

    expected_df = df.copy()
    expected_df["prediction"] = [0.5, 0.5, 0.5, 0.5, 0.5, 0.5]

    predictor_mock = mock.Mock()
    predictor_mock.target = "__target"
    predictor_mock.columns_added = ['prediction']
    predictor_mock.add_prediction.return_value = expected_df
    predictor_mock._estimator_features = [RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED]

    column_names = ColumnNames(
        match_id="game_id",
        team_id="team_id",
        player_id="player_id",
        start_date="start_date",
    )

    lag_transformer = LagTransformer(features=["kills", "deaths"], lag_length=1, granularity=['player_id'],
                                     prefix='lag_')

    pipeline = Pipeline(
        rating_generators=[],
        post_rating_transformers=[
            lag_transformer],
        predictor=predictor_mock,
        column_names=column_names
    )

    new_df = pipeline.train_predict(df=df)

    pd.testing.assert_frame_equal(new_df, expected_df, check_like=True)

    col_names_predictor_train = predictor_mock.train.call_args[1]['df'].columns.tolist()
    assert any(lag_transformer.prefix in element for element in col_names_predictor_train)

    col_names_predictor_add = predictor_mock.add_prediction.call_args[1]['df'].columns.tolist()
    assert any(lag_transformer.prefix in element for element in col_names_predictor_add)


def test_match_predictor_generate_and_predict():
    historical_df = pd.DataFrame({
        "game_id": [1, 1, 2, 2, 3, 3],
        "player_id": [1, 2, 3, 1, 2, 3],
        "team_id": [1, 2, 1, 2, 1, 3],
        "start_date": [pd.to_datetime("2023-01-01"), pd.to_datetime("2023-01-01"), pd.to_datetime("2023-01-02"),
                       pd.to_datetime("2023-01-02"), pd.to_datetime("2023-01-03"), pd.to_datetime("2023-01-03")],
        'deaths': [1, 1, 1, 2, 2, 2],
        "kills": [0.2, 0.3, 0.4, 0.5, 2, 0.2],
        "__target": [1, 0, 1, 0, 1, 0],
    })

    future_df = pd.DataFrame(
        {
            "game_id": [4, 4, 5, 5],
            "player_id": [1, 2, 1, 3],
            "team_id": [1, 3, 1, 3],
            "start_date": [pd.to_datetime("2023-01-04"), pd.to_datetime("2023-01-04"), pd.to_datetime("2023-01-05"),
                           pd.to_datetime("2023-01-05")],
        }
    )

    expected_future_df = future_df.copy()
    expected_future_df["prediction"] = [0.5, 0.5, 0.5, 0.5]

    column_weights = [
        ColumnWeight(name="kills", weight=0.6),
        ColumnWeight(name="deaths", weight=0.4, lower_is_better=True)
    ]

    historical_df_mock_return_with_prediction = historical_df.copy()
    historical_df_mock_return_with_prediction["prediction"] = [0.5, 0.5, 0.5, 0.5, 0.5, 0.5]

    predictor_mock = mock.Mock()
    predictor_mock.target = "__target"
    predictor_mock.columns_added = ['prediction']
    predictor_mock.add_prediction.side_effect = [historical_df_mock_return_with_prediction, expected_future_df]
    predictor_mock._estimator_features = [RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED]

    column_names = ColumnNames(
        match_id="game_id",
        team_id="team_id",
        player_id="player_id",
        start_date="start_date",
    )
    rating_generator = UpdateRatingGenerator(
        estimator_features_out=[RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED])

    pipeline = Pipeline(
        performances_generator=PerformancesGenerator(Performance(weights=column_weights)),
        predictor=predictor_mock,
        rating_generators=rating_generator,
        column_names=column_names
    )

    _ = pipeline.train_predict(df=historical_df)
    new_df = pipeline.future_predict(future_df)

    pd.testing.assert_frame_equal(new_df, expected_future_df, check_like=True)

    assert len(pipeline.performances_generator.pre_transformations) > 0


def test_train_predict_cross_validate():
    historical_df = pd.DataFrame({
        "game_id": [1, 1, 2, 2, 3, 3, 4, 4, 5, 5],
        "player_id": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
        "team_id": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
        "start_date": [pd.to_datetime("2023-01-01"), pd.to_datetime("2023-01-01"), pd.to_datetime("2023-01-02"),
                       pd.to_datetime("2023-01-02"), pd.to_datetime("2023-01-03"), pd.to_datetime("2023-01-03"),
                       pd.to_datetime("2023-01-03"), pd.to_datetime("2023-01-03"),
                       pd.to_datetime("2023-01-04"), pd.to_datetime("2023-01-04"),
                       ],
        'deaths': [1, 1, 1, 2, 2, 2, 1, 3, 2, 5],
        "kills": [0.2, 0.3, 0.4, 0.5, 2, 0.2, 1, 4, 3, 5],
        "__target": [1, 0, 1, 0, 1, 0, 1, 0, 1, 0],
    })

    column_weights = [
        ColumnWeight(name="kills", weight=0.6),
        ColumnWeight(name="deaths", weight=0.4, lower_is_better=True)
    ]
    predictor = Predictor(estimator=LinearRegression())

    column_names = ColumnNames(
        match_id="game_id",
        team_id="team_id",
        player_id="player_id",
        start_date="start_date",
    )
    rating_generator = UpdateRatingGenerator(
        estimator_features_out=[RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED])

    pipeline = Pipeline(
        performances_generator=PerformancesGenerator(Performance(weights=column_weights)),
        predictor=predictor,
        rating_generators=rating_generator,
        post_rating_transformers=[
            LagTransformer(features=["kills", "deaths"], lag_length=1, granularity=['player_id'])],
        column_names=column_names
    )

    cross_validated_df = pipeline.train_predict(df=historical_df, cross_validate_predict=True)


def test_cross_validate_is_equal_to_predict_future():
    df = pd.DataFrame({
        "game_id": [1, 1, 2, 2, 3, 3, 4, 4, 5, 5],
        "player_id": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
        "team_id": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
        "start_date": [pd.to_datetime("2023-01-01"), pd.to_datetime("2023-01-01"), pd.to_datetime("2023-01-02"),
                       pd.to_datetime("2023-01-02"), pd.to_datetime("2023-01-03"), pd.to_datetime("2023-01-03"),
                       pd.to_datetime("2023-01-03"), pd.to_datetime("2023-01-03"),
                       pd.to_datetime("2023-01-04"), pd.to_datetime("2023-01-04"),
                       ],
        'deaths': [0.5, 1, 0.7, 2, 0.5, 0.7, 0.2, 2.1, 0.8, 1],
        "kills": [0.2, 0.3, 0.2, 0.3,0.2, 0.3, 0.2, 0.3, 0.2, 0.3],
        "__target": [1, 0, 0.6, 0.3, 0.8, 0.2, 0.4, 0.1, 1, 0],
    })

    column_weights = [
        ColumnWeight(name="kills", weight=1)
    ]
    predictor = Predictor(estimator=LinearRegression())

    column_names = ColumnNames(
        match_id="game_id",
        team_id="team_id",
        player_id="player_id",
        start_date="start_date",
    )
    rating_generator = UpdateRatingGenerator(
        estimator_features_out=[RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED])

    pipeline = Pipeline(
        performances_generator=PerformancesGenerator(Performance(weights=column_weights)),
        predictor=predictor,
        rating_generators=rating_generator,
        post_rating_transformers=[
            LagTransformer(features=["kills", "deaths"], lag_length=1, granularity=['player_id'])],
        column_names=column_names,
    )

    cross_validator = MatchCountCrossValidator(
                                               match_id_column_name=column_names.match_id,
                                               validation_match_count=1, n_splits=1)
    cross_validated_df = pipeline.cross_validate_predict(df=df, cross_validator=cross_validator, add_train_prediction=True)

    historical_df = df[df[column_names.start_date] < pd.to_datetime("2023-01-04")]
    future_df = df[df[column_names.start_date] >= pd.to_datetime("2023-01-04")]
    historical_df_predictions = pipeline.train_predict(df=historical_df)
    future_predict = pipeline.future_predict(df=future_df).reset_index(drop=True)
    future_cv_df = cross_validated_df[cross_validated_df[column_names.start_date] >= pd.to_datetime("2023-01-04")].reset_index(drop=True)
    past_cv_df = cross_validated_df[cross_validated_df[column_names.start_date] < pd.to_datetime("2023-01-04")].reset_index(drop=True)
    pd.testing.assert_frame_equal(future_cv_df, future_predict, check_like=True, check_dtype=False)
    pd.testing.assert_frame_equal(past_cv_df, historical_df_predictions, check_like=True, check_dtype=False)

def test_train_predict_cross_validate_is_equal_to_predict_future():
    df = pd.DataFrame({
        "game_id": [1, 1, 2, 2, 3, 3, 4, 4, 5, 5],
        "player_id": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
        "team_id": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
        "start_date": [pd.to_datetime("2023-01-01"), pd.to_datetime("2023-01-01"), pd.to_datetime("2023-01-02"),
                       pd.to_datetime("2023-01-02"), pd.to_datetime("2023-01-03"), pd.to_datetime("2023-01-03"),
                       pd.to_datetime("2023-01-03"), pd.to_datetime("2023-01-03"),
                       pd.to_datetime("2023-01-04"), pd.to_datetime("2023-01-04"),
                       ],
        'deaths': [0.5, 1, 0.7, 2, 0.5, 0.7, 0.2, 2.1, 0.8, 1],
        "kills": [0.2, 0.3, 0.2, 0.3,0.2, 0.3, 0.2, 0.3, 0.2, 0.3],
        "__target": [1, 0, 0.6, 0.3, 0.8, 0.2, 0.4, 0.1, 1, 0],
    })

    column_weights = [
        ColumnWeight(name="kills", weight=1)
    ]
    predictor = Predictor(estimator=LinearRegression())

    column_names = ColumnNames(
        match_id="game_id",
        team_id="team_id",
        player_id="player_id",
        start_date="start_date",
    )
    rating_generator = UpdateRatingGenerator(
        estimator_features_out=[RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED])

    pipeline = Pipeline(
        performances_generator=PerformancesGenerator(Performance(weights=column_weights)),
        predictor=predictor,
        rating_generators=rating_generator,
        post_rating_transformers=[
            LagTransformer(features=["kills", "deaths"], lag_length=1, granularity=['player_id'])],
        column_names=column_names,
    )

    cross_validator = MatchCountCrossValidator(
                                               match_id_column_name=column_names.match_id,
                                               validation_match_count=1, n_splits=1)
    cross_validated_df = pipeline.train_predict(df=df, cross_validate_predict=True, cross_validator=cross_validator)

    historical_df = df[df[column_names.start_date] < pd.to_datetime("2023-01-04")]
    future_df = df[df[column_names.start_date] >= pd.to_datetime("2023-01-04")]
    historical_df_predictions = pipeline.train_predict(df=historical_df)
    future_predict = pipeline.future_predict(df=future_df).reset_index(drop=True)
    future_cv_df = cross_validated_df[cross_validated_df[column_names.start_date] >= pd.to_datetime("2023-01-04")].reset_index(drop=True)
    past_cv_df = cross_validated_df[cross_validated_df[column_names.start_date] < pd.to_datetime("2023-01-04")].reset_index(drop=True)
    pd.testing.assert_frame_equal(future_cv_df, future_predict, check_like=True, check_dtype=False)
    pd.testing.assert_frame_equal(past_cv_df, historical_df_predictions, check_like=True, check_dtype=False)


def test_cross_validation_does_not_mutate():
    pass



import pandas as pd

from player_performance_ratings import ColumnNames


def validate_sorting(df: pd.DataFrame, column_names: ColumnNames) -> None:
    df_sorted = df.sort_values(
        by=[column_names.start_date, column_names.match_id,
            column_names.team_id, column_names.player_id])

    if not df.equals(df_sorted):
        for column in [column_names.match_id, column_names.team_id, column_names.player_id]:
            df = df.assign(**{column: df[column].astype('str')})

        df_sorted = df.sort_values(
            by=[column_names.start_date, column_names.match_id,
                column_names.team_id, column_names.player_id])

        if df.equals(df_sorted):
            return
        for column in [column_names.match_id, column_names.team_id, column_names.player_id]:
            try:
                df = df.assign(**{column: df[column].astype('int')})
            except:
                pass

        df_sorted = df.sort_values(
            by=[column_names.start_date, column_names.match_id,
                column_names.team_id, column_names.player_id])

        if df.equals(df_sorted):
            return

        raise ValueError("df needs to be sorted by date, game_id, team_id, player_id in ascending order")





import copy
import inspect
from typing import Optional

import optuna
import pandas as pd
from optuna.samplers import TPESampler
from optuna.trial import BaseTrial
from player_performance_ratings.transformation.base_transformer import BasePostTransformer

from player_performance_ratings.cross_validator.cross_validator import CrossValidator
from player_performance_ratings import PipelineFactory, ColumnNames

from player_performance_ratings.predictor import BasePredictor


from player_performance_ratings.tuner.utils import ParameterSearchRange, add_params_from_search_range, \
    get_default_lgbm_regressor_search_range


class PredictorTuner():

    def __init__(self,
                 search_ranges: Optional[list[ParameterSearchRange]] = None,
                 default_params: Optional[dict] = None,
                 n_trials: int = 30
                 ):
        self.search_ranges = search_ranges
        self.default_params = default_params or {}
        self.n_trials = n_trials

    def tune(self, df: pd.DataFrame,
             pipeline_factory: PipelineFactory,
             cross_validator: CrossValidator,
             best_post_rating_transformers: Optional[list[BasePostTransformer]] = None,
             ) -> BasePredictor:

        deepest_estimator = pipeline_factory.predictor.estimator

        estimator_subclass_level = 0
        while hasattr(deepest_estimator, "estimator"):
            deepest_estimator = deepest_estimator.estimator
            estimator_subclass_level += 1

        if self.search_ranges is None and deepest_estimator.__class__.__name__ in ('LGBMRegressor', 'LGBMClassifier'):
            search_ranges = get_default_lgbm_regressor_search_range()
        elif self.search_ranges is None:
            raise ValueError("search_ranges can't be None if estimator is not LGBMRegressor or LGBMClassifier")
        else:
            search_ranges = self.search_ranges

        def objective(trial: BaseTrial, df: pd.DataFrame) -> float:

            predictor = pipeline_factory.predictor

            if estimator_subclass_level == 0:
                param_names = list(
                    inspect.signature(predictor.estimator.__class__.__init__).parameters.keys())[1:]
                params = {attr: getattr(predictor.estimator, attr) for attr in param_names if attr != 'kwargs'}
                if '_other_params' in predictor.estimator.__dict__:
                    params.update(predictor.estimator._other_params)
            elif estimator_subclass_level == 1:
                param_names = list(
                    inspect.signature(predictor.estimator.estimator.__class__.__init__).parameters.keys())[1:]
                params = {attr: getattr(predictor.estimator.estimator, attr) for attr in param_names if
                          attr != 'kwargs'}
                if '_other_params' in predictor.estimator.estimator.__dict__:
                    params.update(predictor.estimator.estimator._other_params)
            elif estimator_subclass_level == 2:
                param_names = list(
                    inspect.signature(predictor.estimator.estimator.estimator.__class__.__init__).parameters.keys())[1:]
                params = {attr: getattr(predictor.estimator.estimator.estimator, attr) for attr in param_names if
                          attr != 'kwargs'}
                if '_other_params' in predictor.estimator.estimator.estimator.__dict__:
                    params.update(predictor.estimator.estimator.estimator._other_params)

            else:
                raise ValueError(
                    f"estimator_subclass_level can't be higher than 2, got {estimator_subclass_level}")

            params = add_params_from_search_range(params=params,
                                                  trial=trial,
                                                  parameter_search_range=search_ranges)
            for param, value in self.default_params.items():
                params[param] = value

            predictor = copy.deepcopy(pipeline_factory.predictor)
            for param in params:
                if estimator_subclass_level == 1:
                    setattr(predictor.estimator.estimator, param, params[param])
                elif estimator_subclass_level == 2:
                    setattr(predictor.estimator.estimator.estimator, param, params[param])
                elif estimator_subclass_level > 2:
                    raise ValueError(
                        f"estimator_subclass_level can't be higher than 2, got {estimator_subclass_level}")
                else:
                    setattr(predictor.estimator, param, params[param])

            pipeline = pipeline_factory.create(predictor=predictor,
                                               post_rating_transformers=best_post_rating_transformers)
            create_rating_features = False
            for rating_generator in pipeline.rating_generators:
                create_rating_features = any(feature not in df.columns for feature in  rating_generator.estimator_features_return)
                if create_rating_features:
                    break
            return pipeline.cross_validate_score(df=df, create_performance=False, create_rating_features=create_rating_features,
                                                 cross_validator=cross_validator)

        direction = "minimize"
        study_name = "optuna_study"
        optuna_seed = 12
        sampler = TPESampler(seed=optuna_seed)
        study = optuna.create_study(direction=direction, study_name=study_name, sampler=sampler)
        callbacks = []
        study.optimize(lambda trial: objective(trial, df), n_trials=self.n_trials, callbacks=callbacks)
        best_estimator_params = study.best_params
        other_predictor_params = list(
            inspect.signature(pipeline_factory.predictor.__class__.__init__).parameters.keys())[1:]

        if estimator_subclass_level > 0:
            if estimator_subclass_level == 1:
                best_estimator_params.update(pipeline_factory.predictor.estimator.estimator._other_params)
            elif estimator_subclass_level == 2:
                best_estimator_params.update(
                    pipeline_factory.predictor.estimator.estimator.estimator._other_params)

        else:
            if '_other_params' in pipeline_factory.predictor.estimator.__dict__:
                best_estimator_params.update(pipeline_factory.predictor.estimator._other_params)

        other_predictor_params = {attr: getattr(pipeline_factory.predictor, attr) for attr in
                                  other_predictor_params if attr not in ('estimator')}

        predictor_class = pipeline_factory.predictor.__class__
        if estimator_subclass_level == 1:

            potential_parent_names = list(
                inspect.signature(
                    pipeline_factory.predictor.estimator.__class__.__init__).parameters.keys())[1:]
            other_parent_params = {attr: getattr(pipeline_factory.predictor.estimator, attr) for attr
                                   in
                                   potential_parent_names if attr != 'estimator'}

            estimator_class = pipeline_factory.predictor.estimator.estimator.__class__
            parent_estimator_class = pipeline_factory.predictor.estimator.__class__
            parent_estimator = parent_estimator_class(estimator=estimator_class(**best_estimator_params),
                                                      **other_parent_params)
            return predictor_class(estimator=parent_estimator, **other_predictor_params)
        elif estimator_subclass_level == 2:
            potential_parent_names = list(
                inspect.signature(
                    pipeline_factory.predictor.estimator.estimator.__class__.__init__).parameters.keys())[1:]
            other_parent_params = {attr: getattr(pipeline_factory.predictor.estimator.estimator, attr) for attr
                                   in
                                   potential_parent_names if attr != 'estimator'}
            estimator_class = pipeline_factory.predictor.estimator.estimator.estimator.__class__
            parent_estimator_class = pipeline_factory.predictor.estimator.estimator.__class__
            parent_estimator = parent_estimator_class(estimator=estimator_class(**best_estimator_params),
                                                      **other_parent_params)

            parent_parent_estimator_class = pipeline_factory.predictor.estimator.__class__
            parent_parent_estimator = parent_parent_estimator_class(estimator=parent_estimator)
            return predictor_class(estimator=parent_parent_estimator, **other_predictor_params)

        elif estimator_subclass_level == 0:

            estimator_class = pipeline_factory.predictor.estimator.__class__
            return predictor_class(estimator=estimator_class(**best_estimator_params), **other_predictor_params)

        else:
            raise ValueError(f"estimator_subclass_level can't be higher than 2, got {estimator_subclass_level}")



from abc import abstractmethod, ABC
from typing import Optional

import pandas as pd
from player_performance_ratings import ColumnNames


class BaseTransformer(ABC):

    def __init__(self, features: list[str]):
        self.features = features

    @abstractmethod
    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @abstractmethod
    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @property
    @abstractmethod
    def features_out(self) -> list[str]:
        pass


class BasePostTransformer(ABC):

    def __init__(self, features: list[str], are_estimator_features: bool = True):
        self.features = features
        self._are_estimator_features = are_estimator_features
        self._features_out = []
        self.column_names = None

    @abstractmethod
    def fit_transform(self, df: pd.DataFrame, column_names: ColumnNames) -> pd.DataFrame:
        pass

    @abstractmethod
    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @property
    def features_out(self) -> list[str]:
        return self._features_out

    @property
    def estimator_features_out(self) -> list[str]:
        if self._are_estimator_features:
            return self.features_out
        return []


class BaseLagTransformer(BasePostTransformer):

    def __init__(self,
                 features: list[str],
                 add_opponent: bool,
                 iterations: list[int],
                 prefix: str,
                 are_estimator_features: bool = True,
                 ):
        super().__init__(features, are_estimator_features)
        self._entity_features = []
        self.add_opponent = add_opponent
        self.prefix = prefix
        self._df = None
        self._entity_features = []

        for feature_name in self.features:
            for lag in iterations:
                self._features_out.append(f'{prefix}{lag}_{feature_name}')
                self._entity_features.append(f'{prefix}{lag}_{feature_name}')
                if self.add_opponent:
                    self._features_out.append(f'{prefix}{lag}_{feature_name}_opponent')

    def _concat_df(self, df: pd.DataFrame) -> pd.DataFrame:
        df = self._string_convert(df=df)
        df = df.assign(
            __id=df[[self.column_names.match_id, self.column_names.team_id,
                     self.column_names.player_id]].agg('__'.join, axis=1))

        for feature in self.features:
            if feature in df.columns:
                df = df.assign(**{feature: lambda x: x[feature].astype('float')})

        concat_df = pd.concat([self._df, df], axis=0).reset_index()
        if concat_df[self.column_names.start_date].dtype in ('str', 'object'):
            concat_df[self.column_names.start_date] = pd.to_datetime(concat_df[self.column_names.start_date])
        return concat_df.drop_duplicates(subset=['__id'], keep='last')

    def _fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = self._string_convert(df)
        if self._df is None:
            self._df = df
        else:
            self._df = pd.concat([self._df, df], axis=0)

        self._df = self._df.assign(
            __id=self._df[[self.column_names.match_id, self.column_names.team_id,
                           self.column_names.player_id]].agg('__'.join, axis=1))
        self._df = self._df.drop_duplicates(subset=['__id'], keep='last')

        transformed_df = self.transform(df=df)
        return transformed_df

    def _string_convert(self, df: pd.DataFrame) -> pd.DataFrame:
        for column in [self.column_names.match_id, self.column_names.parent_team_id, self.column_names.player_id,
                       self.column_names.update_match_id]:
            df = df.assign(**{column: lambda x: x[column].astype('str')})
        return df

    def _create_transformed_df(self, df: pd.DataFrame, concat_df: pd.DataFrame) -> pd.DataFrame:

        if self.add_opponent:
            concat_df = self._add_opponent_features(df=concat_df)

        ori_cols = df.columns.tolist()
        ori_index_values = df.index.tolist()

        df = self._string_convert(df)
        df = df.assign(
            __id=df[[self.column_names.match_id, self.column_names.team_id,
                     self.column_names.player_id]].agg('__'.join, axis=1))

        transformed_df = concat_df[concat_df['__id'].isin(df['__id'].unique().tolist())][ori_cols + self._features_out]
        transformed_df.index = ori_index_values
        transformed_df = transformed_df.sort_values(by=[self.column_names.start_date, self.column_names.match_id,
                                                        self.column_names.team_id, self.column_names.player_id])
        return transformed_df[list(set(ori_cols + self._features_out))]

    def _add_opponent_features(self, df: pd.DataFrame) -> pd.DataFrame:
        team_features = df.groupby([self.column_names.team_id, self.column_names.match_id])[
            self._entity_features].mean().reset_index()
        df_opponent_feature = team_features.rename(
            columns={**{self.column_names.team_id: '__opponent_team_id'},
                     **{f: f"{f}_opponent" for f in self._entity_features}}
        )
        new_df = df.merge(df_opponent_feature, on=[self.column_names.match_id], suffixes=('', '_team_sum'))
        new_df = new_df[new_df[self.column_names.team_id] != new_df['__opponent_team_id']].drop(
            columns=['__opponent_team_id'])

        new_feats = [f"{f}_opponent" for f in self._entity_features]
        return df.merge(
            new_df[[self.column_names.match_id, self.column_names.team_id, self.column_names.player_id,
                    *new_feats]],
            on=[self.column_names.match_id, self.column_names.team_id, self.column_names.player_id], how='left')



import logging
from typing import List, Optional, Union

import pandas as pd
from sklearn.metrics import log_loss, mean_absolute_error

from player_performance_ratings.scorer import SklearnScorer, OrdinalLossScorer

from player_performance_ratings.cross_validator.cross_validator import CrossValidator, MatchKFoldCrossValidator
from player_performance_ratings.ratings import PerformancesGenerator

from player_performance_ratings.consts import PredictColumnNames
from player_performance_ratings.predictor._base import BasePredictor

from player_performance_ratings.data_structures import Match, ColumnNames
from player_performance_ratings.ratings.league_identifier import LeagueIdentifier
from player_performance_ratings.ratings.match_generator import convert_df_to_matches
from player_performance_ratings.ratings.rating_generator import RatingGenerator

from player_performance_ratings.transformation.base_transformer import BasePostTransformer


class Pipeline():

    def __init__(self,
                 predictor: BasePredictor,
                 column_names: ColumnNames,
                 rating_generators: Optional[Union[RatingGenerator, list[RatingGenerator]]] = None,
                 performances_generator: Optional[PerformancesGenerator] = None,
                 post_rating_transformers: Optional[List[BasePostTransformer]] = None,
                 ):

        """

        :param column_names:
        :param rating_generators:
        A single or a list of RatingGenerators.

        :param performances_generator:
        An optional transformer class that take place in order to convert one or multiple column names into the performance value that is used by the rating model

        :param post_rating_transformers:
            After rating-generation, additional feature engineering can be performed.

        :param predictor:
            The object which trains and returns predictions. Defaults to LGBMClassifier

        :param estimator: Sklearn-like estimator. If predictor is set, estimator will be ignored.
         Because it sometimes can be tricky to identify the names of all the features that must be passed to predictor, the user can decide to only pass in an estimator.
         The features will then be automatically identified based on features_created from the rating_generator, post_rating_transformers and other_features.
         If predictor is set, estimator will be ignored

        :param other_features: If estimator is set and predictor is not,
        other_features allows the user to pass in additional features that are not created by the rating_generator or post_rating_transformers to the predictor.

        :param other_categorical_features: Which of the other_features are categorical.
        It is not required to duplicate the categorical features in other_features and other_categorical_features.
        Simply passing an a categorical_feature in categorical_features will add it to other_features if it doesn't already exist.

        :param train_split_date:
            Date threshold which defines which periods to train the predictor on

        :param date_column_name:
            If rating_generators are not defined and train_split_date is used, then train_column_name must be set.

        :param use_auto_create_performance_calculator:
            If true, the pre_rating_transformers will be automatically generated to ensure the performance-value is done according to good practices.
            For new users, this is recommended.


        """

        self._estimator_features = predictor._estimator_features
        self.rating_generators: list[RatingGenerator] = rating_generators if isinstance(rating_generators, list) else [
            rating_generators]
        if rating_generators is None:
            self.rating_generators: list[RatingGenerator] = []

        self.post_rating_transformers = post_rating_transformers or []
        self.column_names = column_names

        for c in self.post_rating_transformers:
            self._estimator_features += [f for f in c.estimator_features_out if f not in self._estimator_features]
        for rating_idx, c in enumerate(self.rating_generators):
            for rating_feature in c.estimator_features_out:
                if len(self.rating_generators) > 1:
                    rating_feature_str = rating_feature + str(rating_idx)

                else:
                    rating_feature_str = rating_feature
                if rating_feature_str not in self._estimator_features:
                    self._estimator_features.append(rating_feature_str)

        self.performances_generator = performances_generator
        self.predictor = predictor
        self.predictor.set_target(PredictColumnNames.TARGET)

    def cross_validate_score(self,
                             df: pd.DataFrame,
                             cross_validator: Optional[CrossValidator] = None,
                             matches: Optional[list[Match]] = None,
                             create_performance: bool = True,
                             create_rating_features: bool = True) -> float:

        for col in self.predictor.columns_added:
            if col in df.columns:
                df = df.drop(columns=[col])

        if cross_validator is None:
            cross_validator = self.create_default_cross_validator(df=df)

        if create_performance:
            df = self._add_performance(df=df)

        for rating_generator in self.rating_generators:
            create_rating_features = any(
                feature not in df.columns for feature in rating_generator.estimator_features_return)
            if create_rating_features:
                break

        if create_rating_features and self.rating_generators:
            df = self._add_rating(matches=matches, df=df, store_ratings=False)

        validation_df = cross_validator.generate_validation_df(df=df, predictor=self.predictor,
                                                               column_names=self.column_names,
                                                               post_transformers=self.post_rating_transformers,
                                                               estimator_features=self._estimator_features,
                                                               return_features=False)

        if cross_validator.scorer is None:
            scorer = self._create_default_scorer(df)
            return cross_validator.cross_validation_score(validation_df=validation_df, scorer=scorer)

        return cross_validator.cross_validation_score(validation_df=validation_df)

    def cross_validate_predict(self,
                               df: pd.DataFrame,
                               cross_validator: Optional[CrossValidator] = None,
                               matches: Optional[list[Match]] = None,
                               create_performance: bool = True,
                               create_rating_features: bool = True,
                               return_features: bool = False,
                               add_train_prediction: bool = False
                               ) -> pd.DataFrame:


        cross_validated_df = df.copy()
        if cross_validator is None:
            cross_validator = self.create_default_cross_validator(df=cross_validated_df)

        if self.predictor.target not in cross_validated_df.columns:
            raise ValueError(
                f"Target {self.predictor.target} not in df columns. Target always needs to be set equal to {PredictColumnNames.TARGET}")

        if create_performance:
            cross_validated_df = self._add_performance(df=cross_validated_df)

        for rating_generator in self.rating_generators:
            create_rating_features = any(
                feature not in df.columns for feature in rating_generator.estimator_features_return)
            if create_rating_features:
                break

        if create_rating_features and self.rating_generators:
            cross_validated_df = self._add_rating(matches=matches, df=cross_validated_df, store_ratings=False)

        cross_validated_df = cross_validator.generate_validation_df(df=cross_validated_df,
                                                                    predictor=self.predictor,
                                                                    column_names=self.column_names,
                                                                    post_transformers=self.post_rating_transformers,
                                                                    estimator_features=self._estimator_features,
                                                                    return_features=return_features,
                                                                    add_train_prediction=add_train_prediction)

        cn = self.column_names
        for _, row in df[[cn.match_id, cn.team_id, cn.player_id]].dtypes.reset_index().iterrows():
            cross_validated_df[row['index']] = cross_validated_df[row['index']].astype(row[0])
        if return_features:
            cols_to_drop = []
            for c in list(set(self._estimator_features + self.predictor.columns_added)):
                if c in cross_validated_df.columns and c in df.columns:
                    cols_to_drop.append(c)
            df = df.drop(columns=cols_to_drop)
            new_feats = [f for f in cross_validated_df.columns if f not in df.columns]
            return df.merge(
                cross_validated_df[new_feats + [cn.match_id, cn.team_id, cn.player_id]],
                on=[cn.match_id, cn.team_id, cn.player_id], how='left')

        predictor_cols_added = self.predictor.columns_added
        if 'classes' in cross_validated_df.columns and 'classes' not in predictor_cols_added and 'classes' not in df.columns:
            predictor_cols_added.append('classes')

        return df.merge(
            cross_validated_df[predictor_cols_added + [cn.match_id, cn.team_id, cn.player_id]],
            on=[cn.match_id, cn.team_id, cn.player_id], how='left')

    def create_default_cross_validator(self, df: pd.DataFrame) -> CrossValidator:

        scorer = self._create_default_scorer(df)

        return MatchKFoldCrossValidator(
            date_column_name=self.column_names.start_date,
            match_id_column_name=self.column_names.update_match_id,
            scorer=scorer,
        )

    def _create_default_scorer(self, df: pd.DataFrame):
        if self.predictor.estimator_type == "regressor":
            scorer = SklearnScorer(scorer_function=mean_absolute_error, pred_column=self.predictor.pred_column)
            logging.info("Using mean_absolute_error as scorer")
        else:
            if len(df[PredictColumnNames.TARGET].unique()) > 2:
                scorer = OrdinalLossScorer(pred_column=self.predictor.pred_column)
                logging.info("Using ordinal loss as scorer")
            else:
                scorer = SklearnScorer(scorer_function=log_loss, pred_column=self.predictor.pred_column)
                logging.info("Using log_loss as scorer")

        return scorer

    def train_predict(self,
                      df: pd.DataFrame,
                      matches: Optional[Union[list[Match], list[list[Match]]]] = None,
                      store_ratings: bool = True,
                      return_features: bool = False,
                      cross_validate_predict: bool = False,
                      cross_validator: Optional[CrossValidator] = None,
                      ) -> pd.DataFrame:

        self.reset_pipeline()
        df_with_predict = df.copy()

        if self.predictor.target not in df_with_predict.columns:
            raise ValueError(
                f"Target {self.predictor.target} not in df columns. Target always needs to be set equal to {PredictColumnNames.TARGET}")

        ori_cols = df.columns.tolist()
        df_with_predict = self._add_performance(df=df_with_predict)
        if self.rating_generators:
            df_with_predict = self._add_rating(matches=matches, df=df_with_predict, store_ratings=store_ratings)

        if cross_validate_predict:
            df_cv_predict = self.cross_validate_predict(df=df_with_predict, return_features=return_features,
                                                          create_rating_features=False, create_performance=False,
                                                          add_train_prediction=True, cross_validator=cross_validator)

        for post_rating_transformer in self.post_rating_transformers:
            df_with_predict = post_rating_transformer.fit_transform(df_with_predict, column_names=self.column_names)

        self.predictor.train(df=df_with_predict, estimator_features=self._estimator_features)

        if cross_validate_predict:
            df_with_predict = df_cv_predict
        else:
            df_with_predict = self.predictor.add_prediction(df=df_with_predict)
        cn = self.column_names
        for _, row in df[[cn.match_id, cn.team_id, cn.player_id]].dtypes.reset_index().iterrows():
            df_with_predict[row['index']] = df_with_predict[row['index']].astype(row[0])

        if return_features:
            new_feats = [f for f in df_with_predict.columns if f not in ori_cols]
            return df.merge(
                df_with_predict[new_feats + [cn.match_id, cn.team_id, cn.player_id]],
                on=[cn.match_id, cn.team_id, cn.player_id], how='left')

        predictor_cols_added = self.predictor.columns_added
        if 'classes' in df_with_predict.columns and 'classes' not in predictor_cols_added and 'classes' not in df.columns:
            predictor_cols_added.append('classes')

        return df.merge(
            df_with_predict[predictor_cols_added + [cn.match_id, cn.team_id, cn.player_id]],
            on=[cn.match_id, cn.team_id, cn.player_id], how='left')

    def reset_pipeline(self):
        for idx in range(len(self.rating_generators)):
            self.rating_generators[idx].reset_ratings()

        for idx in range(len(self.post_rating_transformers)):
            if hasattr(self.post_rating_transformers[idx], "_df"):
                self.post_rating_transformers[idx]._df = None

    def _add_performance(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()

        if self.predictor.pred_column in df.columns:
            raise ValueError(
                f"Predictor column {self.predictor.pred_column} already in df columns. Remove or rename before generating predictions")


        elif self.performances_generator:
            df = self.performances_generator.generate(df)

        if self.predictor.target not in df.columns:
            raise ValueError(
                f"Target {self.predictor.target} not in df columns. Target always needs to be set equal to {PredictColumnNames.TARGET}")

        return df

    def _add_rating(self, matches: Optional[Union[list[Match], Match]], df: pd.DataFrame,
                    store_ratings: bool = True):

        if matches:
            if isinstance(matches[0], Match):
                matches = [matches for _ in self.rating_generators]

        rg = self.rating_generators[0]
        ratings_df = rg.ratings_df

        if isinstance(ratings_df, pd.DataFrame):
            rating_game_ids = ratings_df[rg.column_names.match_id].unique()
            new_game_ids = df[rg.column_names.match_id].unique()
        else:
            rating_game_ids = []

        df_no_ratings = df[~df[self.column_names.match_id].isin(rating_game_ids)]
        df_calculated_ratings = df[df[self.column_names.match_id].isin(rating_game_ids)]
        for rating_idx, rating_generator in enumerate(self.rating_generators):
            if len(df_no_ratings) > 0:

                if matches is None:
                    rating_matches = convert_df_to_matches(column_names=self.column_names,
                                                           df=df_no_ratings,
                                                           league_identifier=LeagueIdentifier(),
                                                           performance_column_name=rating_generator.performance_column)
                else:
                    rating_matches = matches[rating_idx]
                    if len(df_no_ratings) != len(df):
                        rating_matches = [m for m in rating_matches if m.id in new_game_ids]

                if store_ratings:
                    match_ratings = rating_generator.generate_historical(matches=rating_matches,
                                                                         column_names=self.column_names, df=df)
                else:
                    match_ratings = rating_generator.generate_historical(df=None, matches=rating_matches,
                                                                         column_names=self.column_names)

                for rating_feature, values in match_ratings.items():
                    if len(self.rating_generators) > 1:
                        rating_feature_str = rating_feature + str(rating_idx)
                    else:
                        rating_feature_str = rating_feature
                    df_no_ratings[rating_feature_str] = values

            if len(df_calculated_ratings) > 0:
                ratings_df = rating_generator.ratings_df
                rating_cols = rating_generator.estimator_features_return + [
                    rating_generator.column_names.match_id, rating_generator.column_names.team_id,
                    rating_generator.column_names.player_id]
                df_calculated_ratings = df_calculated_ratings[
                    [f for f in df_calculated_ratings.columns if
                     f not in rating_generator.estimator_features_return]].merge(
                    ratings_df[rating_cols], on=[
                        rating_generator.column_names.match_id, rating_generator.column_names.team_id,
                        rating_generator.column_names.player_id], how='inner')

        if len(df_no_ratings) != len(df):
            col = rg.column_names
            df = pd.concat([df_calculated_ratings, df_no_ratings])
            df = df.drop_duplicates(subset=[col.match_id, col.team_id, col.player_id])
        else:
            df = df_no_ratings

        return df

    def future_predict(self, df: pd.DataFrame, return_features: bool = False) -> pd.DataFrame:
        df_with_predict = df.copy()

        for rating_idx, rating_generator in enumerate(self.rating_generators):
            if rating_generator.performance_column in df_with_predict.columns:
                df_with_predict = df_with_predict.drop(columns=[rating_generator.performance_column])
            rating_column_names = rating_generator.column_names

            matches = convert_df_to_matches(column_names=rating_column_names, df=df_with_predict,
                                            league_identifier=LeagueIdentifier(),
                                            performance_column_name=rating_generator.performance_column)

            match_ratings = rating_generator.generate_future(matches=matches, df=df_with_predict)
            for rating_feature, values in match_ratings.items():

                if len(self.rating_generators) > 1:
                    rating_feature_str = rating_feature + str(rating_idx)
                else:
                    rating_feature_str = rating_feature
                df_with_predict[rating_feature_str] = values

        for post_rating_transformer in self.post_rating_transformers:
            df_with_predict = post_rating_transformer.transform(df_with_predict)

        df_with_predict = self.predictor.add_prediction(df_with_predict)

        cn = self.column_names
        for _, row in df[[cn.match_id, cn.team_id, cn.player_id]].dtypes.reset_index().iterrows():
            df_with_predict[row['index']] = df_with_predict[row['index']].astype(row[0])
        if return_features:
            new_feats = [f for f in df_with_predict.columns if f not in df.columns]
            return df.merge(
                df_with_predict[new_feats + [cn.match_id, cn.team_id, cn.player_id]],
                on=[cn.match_id, cn.team_id, cn.player_id], how='left')

        return df.merge(
            df_with_predict[self.predictor.columns_added + [cn.match_id, cn.team_id, cn.player_id]],
            on=[cn.match_id, cn.team_id, cn.player_id], how='left')

    @property
    def classes_(self) -> Optional[list[str]]:
        if 'classes_' not in dir(self.predictor.estimator):
            return None
        return self.predictor.estimator.classes_



import logging
import warnings
import pandas as pd
from lightgbm import LGBMClassifier
from pandas.errors import SettingWithCopyWarning

from player_performance_ratings.predictor.sklearn_estimator import OrdinalClassifier
from player_performance_ratings.predictor_transformer import PredictorTransformer
from player_performance_ratings.scorer.score import Filter, apply_filters

warnings.simplefilter(action="ignore", category=SettingWithCopyWarning)

from typing import Optional

from sklearn.linear_model import LogisticRegression

from player_performance_ratings.consts import PredictColumnNames

from player_performance_ratings.data_structures import ColumnNames
from player_performance_ratings.predictor._base import BasePredictor


class GameTeamPredictor(BasePredictor):

    def __init__(self,
                 game_id_colum: str,
                 team_id_column: str,
                 target: Optional[str] = PredictColumnNames.TARGET,
                 estimator: Optional = None,
                 estimator_features: Optional[list[str]] = None,
                 multiclassifier: bool = False,
                 pred_column: Optional[str] = None,
                 pre_transformers: Optional[list[PredictorTransformer]] = None,
                 filters: Optional[list[Filter]] = None
                 ):
        """
        Wrapper for sklearn models that predicts game results.

        The GameTeam Predictor is intended to transform predictions from a lower granularity into a GameTeam level.
        So if input data is at game-player, data is converted to game_team before being trained
        Similar concept if it is at a granularity below game level.


        :param game_id_colum:
        :param team_id_column:
        :param features:
        :param target:
        :param estimator:
        :param multiclassifier:
        :param pred_column:
        """

        self.game_id_colum = game_id_colum
        self.team_id_column = team_id_column
        self._target = target
        self._estimator_features = []

        self.multiclassifier = multiclassifier
        super().__init__(target=self._target, pred_column=pred_column,
                         estimator=estimator or LogisticRegression(), pre_transformers=pre_transformers,
                         estimator_features=estimator_features, filters=filters)

    def train(self, df: pd.DataFrame, estimator_features: list[Optional[str]] = None) -> None:

        if estimator_features is None and self._estimator_features is None:
            raise ValueError("estimator features must either be passed to .train() or injected into constructor")

        self._estimator_features = estimator_features or self._estimator_features
        self._estimator_features = self._estimator_features.copy()
        df = apply_filters(df=df, filters=self.filters)
        df = self.fit_transform_pre_transformers(df=df)
        if len(df[self._target].unique()) > 2 and hasattr(self.estimator, "predict_proba"):
            self.multiclassifier = True
            if self.estimator.__class__.__name__ == 'LogisticRegression':
                self.estimator = OrdinalClassifier(self.estimator)

        if hasattr(self.estimator, "predict_proba"):
            try:
                df = df.assign(**{self._target: df[self._target].astype('int')})
            except Exception:
                pass

        if self._target not in df.columns:
            raise ValueError(f"target {self._target} not in df")

        grouped = self._create_grouped(df)
        self.estimator.fit(grouped[self._estimator_features], grouped[self._target])

    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        """
        Adds prediction to df

        :param df:
        :return: Input df with prediction column
        """

        if hasattr(self.estimator, "predict_proba"):
            try:
                df[self._target] = df[self._target].astype('int')
            except Exception:
                pass
        if not self._estimator_features:
            raise ValueError("estimator_features not set. Please train first")
        df = self.transform_pre_transformers(df=df)
        grouped = self._create_grouped(df)

        if self.multiclassifier:
            grouped[self._pred_column] = self.estimator.predict_proba(grouped[self._estimator_features]).tolist()
            grouped['classes'] = [list(self.estimator.classes_) for _ in range(len(grouped))]
        elif not hasattr(self.estimator, "predict_proba"):
            grouped[self._pred_column] = self.estimator.predict(grouped[self._estimator_features])
        else:
            grouped[self._pred_column] = self.estimator.predict_proba(grouped[self._estimator_features])[:, 1]

        if self.pred_column in df.columns:
            df = df.drop(columns=[self.pred_column])

        if 'classes' in grouped.columns:
            df = df.merge(grouped[[self.game_id_colum, self.team_id_column] + [self._pred_column, 'classes']],
                          on=[self.game_id_colum, self.team_id_column])

        else:
            df = df.merge(grouped[[self.game_id_colum, self.team_id_column] + [self._pred_column]],
                          on=[self.game_id_colum, self.team_id_column])

        return df

    def _create_grouped(self, df: pd.DataFrame) -> pd.DataFrame:

        numeric_features = [feature for feature in self._estimator_features if df[feature].dtype in ['int', 'float']]
        cat_feats = [feature for feature in self._estimator_features if feature not in numeric_features]

        if self._target in df.columns:
            if df[self._target].dtype == 'object':
                df.loc[:, self._target] = df[self._target].astype('int')

            grouped = df.groupby([self.game_id_colum, self.team_id_column]).agg({
                **{feature: 'mean' for feature in numeric_features},
                self._target: 'mean',
            }).reset_index()

        else:
            grouped = df.groupby([self.game_id_colum, self.team_id_column]).agg({
                **{feature: 'mean' for feature in numeric_features}
            }).reset_index()

        if self._target in df.columns and hasattr(self._deepest_estimator, "predict_proba"):
            grouped[self._target] = grouped[self._target].astype('int')

        grouped = grouped.merge(
            df[[self.game_id_colum, self.team_id_column, *cat_feats]].drop_duplicates(
                subset=[self.game_id_colum, self.team_id_column]),
            on=[self.game_id_colum, self.team_id_column], how='inner')

        return grouped


class Predictor(BasePredictor):

    def __init__(self,
                 target: Optional[str] = PredictColumnNames.TARGET,
                 estimator: Optional = None,
                 estimator_features: Optional[list[str]] = None,
                 filters: Optional[list[Filter]] = None,
                 multiclassifier: bool = False,
                 pred_column: Optional[str] = None,
                 column_names: Optional[ColumnNames] = None,
                 pre_transformers: Optional[list[PredictorTransformer]] = None
                 ):
        self._target = target
        self.multiclassifier = multiclassifier
        self.column_names = column_names

        if estimator is None:
            logging.warning(
                "model is not set. Will use LGBMClassifier(max_depth=2, n_estimators=400, learning_rate=0.05)")

        super().__init__(target=self._target, pred_column=pred_column,
                         estimator=estimator or LGBMClassifier(max_depth=2, n_estimators=300, learning_rate=0.05,
                                                               verbose=-100),
                         pre_transformers=pre_transformers, filters=filters, estimator_features=estimator_features)

    def train(self, df: pd.DataFrame, estimator_features: Optional[list[str]] = None) -> None:

        if estimator_features is None and self._estimator_features is None:
            raise ValueError("estimator features must either be passed to .train() or injected into constructor")
        self._estimator_features = estimator_features or self._estimator_features
        self._estimator_features = self._estimator_features.copy()

        filtered_df = apply_filters(df=df, filters=self.filters)
        if hasattr(self.estimator, "predict_proba"):
            try:
                filtered_df[self._target] = filtered_df[self._target].astype('int')
            except Exception:
                pass

        filtered_df = self.fit_transform_pre_transformers(df=filtered_df)

        if not self.multiclassifier and len(filtered_df[self._target].unique()) > 2 and hasattr(self._deepest_estimator,
                                                                                                "predict_proba"):
            self.multiclassifier = True
            if self.estimator.__class__.__name__ == 'LogisticRegression':
                self.estimator = OrdinalClassifier(self.estimator)
            if len(filtered_df[self._target].unique()) > 50:
                logging.warning(
                    f"target has {len(df[self._target].unique())} unique values. This may machine-learning model to not function properly."
                    f" It is recommended to limit max and min values to ensure less than 50 unique targets")

        if hasattr(self._deepest_estimator, "predict_proba"):
            filtered_df = filtered_df.assign(**{self._target: filtered_df[self._target].astype('int')})

        self.estimator.fit(filtered_df[self._estimator_features], filtered_df[self._target])

    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        if not self._estimator_features:
            raise ValueError("estimator_features not set. Please train first")

        if hasattr(self.estimator, "predict_proba"):
            try:
                df[self._target] = df[self._target].astype('int')
            except Exception:
                pass

        df = self.transform_pre_transformers(df=df)
        df = df.copy()
        if self.multiclassifier:
            df[self._pred_column] = self.estimator.predict_proba(
                df[self._estimator_features]).tolist()
            df['classes'] = [list(self.estimator.classes_) for _ in range(len(df))]
            if len(set(df[self.pred_column].iloc[0])) == 2:
                raise ValueError(
                    "Too many unique values in relation to rows in the training dataset causes multiclassifier to not train properly")

        elif not hasattr(self._deepest_estimator, "predict_proba"):
            df[self._pred_column] = self.estimator.predict(df[self._estimator_features])
        else:
            df[self._pred_column] = self.estimator.predict_proba(df[self._estimator_features])[:, 1]
        return df



from __future__ import annotations

import json
import logging
import warnings

from player_performance_ratings.utils import validate_sorting

warnings.simplefilter(action='ignore', category=FutureWarning)

from typing import Optional
import pandas as pd

from player_performance_ratings.data_structures import MatchTeam, ColumnNames, MatchPerformance, \
    MatchPlayer, Match
from player_performance_ratings.ratings.league_identifier import LeagueIdentifier

HOUR_NUMBER_COLUMN_NAME = "hour_number"


def convert_df_to_matches(
        df: pd.DataFrame,
        column_names: ColumnNames,
        performance_column_name: str,
        league_identifier: Optional[LeagueIdentifier] = LeagueIdentifier(),
) -> list[Match]:
    """
    Converts a dataframe to a list of matches.
    Each dataframe row needs to be a unique combination of match_id and player_id.
    The dataframe needs to contain the following columns:
        * player_id
        * team_id
        * match_id
        * start_date
        * performance


    Dataframe needs to be sorted by date, game_id, team_id in ascending order.

    Optionally a column for participation_weight and league can be passed.
    The participation_weight indicates the percentage (as a ratio) of the time played by the player in the match.
    The league column indicates the league of the match.
    If the league_identifier is passed it will identify the league of the match by the players past matches played.
    If not the  league of the match will be equal to the league of the current match
    """

    df = df.copy()

    if column_names.participation_weight is None and column_names.projected_participation_weight is not None:
        raise ValueError("projected_participation_weight column passed but not participation_weight column")
    if performance_column_name in df.columns:
        if max(df[performance_column_name]) > 1.001 or min(df[performance_column_name]) < -0.001:
            raise ValueError("performance column must be between 0 and 1")

        mean_performance = df[performance_column_name].mean()

        if abs(mean_performance - 0.5) > 0.05:
            logging.warning(
                f"mean performance is {mean_performance} which is far from 0.5. It is recommended to do further pre_transformations of the performance column")


        if df[performance_column_name].isnull().any():
            logging.error(
                f"df[{performance_column_name}] contains nan values. Make sure all column_names used in column_weights are imputed beforehand")
            raise ValueError("performance contains nan values")

    validate_sorting(df=df, column_names=column_names)

    col_names = column_names
    df[col_names.start_date] = pd.to_datetime(df[col_names.start_date], format='%Y-%m-%d %H:%M:%S')
    try:
        date_time = df[col_names.start_date].dt.tz_convert('UTC')
    except TypeError:
        date_time = df[col_names.start_date].dt.tz_localize('UTC')
    df[HOUR_NUMBER_COLUMN_NAME] = (date_time - pd.Timestamp("1970-01-01").tz_localize('UTC')) // pd.Timedelta(
        '1h')

    if col_names.league and col_names.league not in df.columns:
        raise ValueError("league column passed but not in dataframe.")

    league_in_df = False
    if col_names.league is not None:
        league_in_df = True

    if col_names.projected_participation_weight:
        if col_names.projected_participation_weight not in df.columns:
            logging.warning(
                "projected_participation_weight column passed but not in dataframe."
                " Will use participation_weight as projected_participation_weight")

    if col_names.team_players_playing_time:
        if col_names.team_players_playing_time not in df.columns:
            raise ValueError("team_players_playing_time column passed but not in dataframe.")

        is_team_players_playing_time = True
    else:
        is_team_players_playing_time = False

    if col_names.opponent_players_playing_time:
        if col_names.opponent_players_playing_time not in df.columns:
            raise ValueError("opponent_players_playing_time column passed but not in dataframe.")

        is_opponent_players_playing_time = True
    else:
        is_opponent_players_playing_time = False

    prev_match_id = None
    prev_update_team_id = None

    if len(df.columns) > 60:
        logging.warning(
            f"Dataframe column count {len(df.columns)} is high. Consider removing unneeded columns to speed up processing itme")
    data_dict = df.to_dict('records')

    matches = []

    prev_team_id = None
    prev_row: Optional[None, pd.Series] = None
    match_teams = []
    match_team_players = []
    team_league_counts = {}

    for row in data_dict:
        match_id = row[col_names.match_id]
        team_id = row[col_names.team_id]
        update_team_id = row[col_names.parent_team_id]
        if team_id != prev_team_id and prev_team_id != None or prev_match_id != match_id and prev_match_id != None:
            match_team = _create_match_team(team_league_counts=team_league_counts, team_id=prev_team_id,
                                            match_team_players=match_team_players, update_team_id=prev_update_team_id)
            match_teams.append(match_team)
            match_team_players = []
            team_league_counts = {}

        if match_id != prev_match_id and prev_match_id != None:
            match = _create_match(league_in_df=league_in_df, row=prev_row, match_teams=match_teams,
                                  column_names=column_names)
            matches.append(match)
            match_teams = []

        participation_weight = 1.0
        if col_names.participation_weight and col_names.participation_weight in row:
            participation_weight = row[col_names.participation_weight]

        player_id = row[col_names.team_id]
        if col_names.player_id is not None:
            player_id = row[col_names.player_id]

        if col_names.position is not None:
            position = row[col_names.position]
        else:
            position = None

        if league_in_df:
            match_league = row[column_names.league]
            if team_id not in team_league_counts:
                team_league_counts[team_id] = {}
            player_league = league_identifier.identify(player_id=player_id,
                                                       league_match=match_league)
            if player_league not in team_league_counts[team_id]:
                team_league_counts[team_id][player_league] = 0

            team_league_counts[team_id][player_league] += 1
        else:
            player_league = None

        if is_team_players_playing_time:
            team_players_playing_time = row[col_names.team_players_playing_time]
            try:
                team_players_playing_time = json.loads(team_players_playing_time)
            except Exception:
                pass
        else:
            team_players_playing_time = {}

        if is_opponent_players_playing_time:
            opponent_players_playing_time = row[col_names.opponent_players_playing_time]
            try:
                opponent_players_playing_time = json.loads(opponent_players_playing_time)
            except Exception:
                pass
        else:
            opponent_players_playing_time = {}

        if col_names.projected_participation_weight and col_names.participation_weight in row:
            projected_participation_weight = row[col_names.projected_participation_weight]
        elif col_names.participation_weight not in row:
            projected_participation_weight = 1
        else:
            projected_participation_weight = participation_weight

        if performance_column_name in row:
            performance = MatchPerformance(
                team_players_playing_time=team_players_playing_time,
                opponent_players_playing_time=opponent_players_playing_time,
                performance_value=row[performance_column_name],
                participation_weight=participation_weight,
                projected_participation_weight=projected_participation_weight
            )

        else:
            performance = MatchPerformance(
                performance_value=None,
                participation_weight=None,
                projected_participation_weight=participation_weight
            )

        others = {}
        if col_names.other_values:
            for other_col in col_names.other_values:
                others[other_col] = row[other_col]

        match_player = MatchPlayer(
            id=player_id,
            league=player_league,
            performance=performance,
            position=position,
            others=others
        )
        match_team_players.append(match_player)

        prev_match_id = match_id
        prev_team_id = team_id
        prev_update_team_id = update_team_id
        prev_row = row

    match_team = _create_match_team(team_league_counts=team_league_counts, team_id=prev_team_id,
                                    update_team_id=prev_update_team_id,
                                    match_team_players=match_team_players)
    match_teams.append(match_team)
    match = _create_match(league_in_df=league_in_df, row=df.iloc[len(df) - 1],
                          match_teams=match_teams, column_names=column_names)
    matches.append(match)

    return matches


def _create_match(league_in_df, row: pd.Series, match_teams: list[MatchTeam], column_names: ColumnNames) -> Match:
    match_id = row[column_names.match_id]
    if league_in_df:
        match_league = row[column_names.league]
    else:
        match_league = None

    return Match(
        id=match_id,
        teams=match_teams,
        day_number=int(row[HOUR_NUMBER_COLUMN_NAME] / 24),
        league=match_league,
        update_id=row[column_names.update_match_id]
    )


def _create_match_team(team_league_counts: dict, team_id: str,
                       update_team_id: str,
                       match_team_players: list[MatchPlayer]) -> MatchTeam:
    if team_league_counts:
        team_league = max(team_league_counts[team_id], key=team_league_counts[team_id].get)
    else:
        team_league = None
    return MatchTeam(
        id=team_id,
        update_id=update_team_id,
        players=match_team_players,
        league=team_league
    )



from .post_transformers import LagTransformer, RollingMeanTransformer, RollingMeanDaysTransformer
from .pre_transformers import MinMaxTransformer, DiminishingValueTransformer, \
    SklearnEstimatorImputer, PartialStandardScaler, NetOverPredictedTransformer, SymmetricDistributionTransformer, \
    GroupByTransformer
